"""Processing module for CERF

@author Chris R. vernon
@email chris.vernon@pnnl.gov

License:  BSD 2-Clause, see LICENSE and DISCLAIMER files

"""

import logging
import os
import time

import pandas as pd
from joblib import Parallel, delayed

import cerf.utils as util
from cerf.model import Model
from cerf.process_state import process_state


def generate_model(config_file=None, config_dict={}, initialize_site_data=None, log_level='info'):
    """Generate model instance for use in parallel applications.

    :param config_file:                 Full path with file name and extension to the input config.yml file
    :type config_file:                  str

    :param config_dict:                 Optional instead of config_file. Configuration dictionary.
    :type config_dict:                  dict

    :param   initialize_site_data:      None if no initialization is required, otherwise either a CSV file or
                                        Pandas DataFrame of siting data bearing the following required fields:

                                        xcoord:  the X coordinate of the site in meters in
                                        USA_Contiguous_Albers_Equal_Area_Conic (EPSG:  102003)

                                        ycoord:  the Y coordinate of the site in meters in
                                        USA_Contiguous_Albers_Equal_Area_Conic (EPSG:  102003)

                                        retirement_year:  the year (int four digit, e.g., 2050) that the power
                                        plant is to be decommissioned

                                        buffer_in_km:  the buffer around the site to apply in kilometers

    :param log_level:                   Log level.  Options are 'info' and 'debug'.  Default 'info'
    :type log_level:                    str

    """

    return Model(config_file, config_dict, initialize_site_data, log_level)


def cerf_parallel(model, data, write_output=True, n_jobs=-1, method='sequential'):
    """Run all CONUS states in parallel.

    :param model:                       Instantiated CERF model class containing configuration options
    :type model:                        class

    :param data:                        Data from cerf.stage.Stage containing NLC and suitability arrays

    :param config_file:                 Full path with file name and extension to the input config.yml file
    :type config_file:                  str

    :param write_output:                Write output as a raster to the output directory specified in the config file
    :type write_output:                 bool

    :param n_jobs:                      The number of processors to utilize.  Default is -1 which is all but 1.
    :type n_jobs:                       int

    :param method:                      Backend parallelization method used in Joblib.  Default is `sequential` to
                                        manage overhead for local runs.  Options for advanced configurations are:
                                        `loky`, `threading`, and `multiprocessing`.
                                        See https://joblib.readthedocs.io/en/latest/parallel.html for details.
    :type method:                       str

    :return:                            A 2D arrays containing sites as the technology ID per grid cell.  All
                                        non-sited grid cells are given the value of NaN.

    """

    # start time for parallel run
    t0 = time.time()

    # run all states in parallel
    results = Parallel(n_jobs=n_jobs, backend=method)(delayed(process_state)(target_state_name=i,
                                                                             settings_dict=model.settings_dict,
                                                                             technology_dict=model.technology_dict,
                                                                             technology_order=model.technology_order,
                                                                             expansion_dict=model.expansion_dict,
                                                                             states_dict=model.states_dict,
                                                                             suitability_arr=data.suitability_arr,
                                                                             lmp_arr=data.lmp_arr,
                                                                             nov_arr=data.nov_arr,
                                                                             ic_arr=data.ic_arr,
                                                                             nlc_arr=data.nlc_arr,
                                                                             zones_arr=data.zones_arr,
                                                                             xcoords=data.xcoords,
                                                                             ycoords=data.ycoords,
                                                                             indices_2d=data.indices_2d,
                                                                             randomize=model.settings_dict.get('randomize', True),
                                                                             seed_value=model.settings_dict.get('seed_value', 0),
                                                                             verbose=model.settings_dict.get('verbose', False),
                                                                             write_output=False) for i in model.states_dict.keys())

    logging.info(f"All states processed in {round((time.time() - t0), 7)} seconds.")
    logging.info("Aggregating outputs...")

    # create a data frame to hold the outputs
    df = pd.DataFrame(util.empty_sited_dict()).astype(util.sited_dtypes())

    # add in the initialized siting data from a previous years run if so desired
    if model.initialize_site_data is not None:
        df = pd.concat([df, data.init_df])

    # combine the outputs for all states
    for i in results:

        # ensure some sites were able to be sited for the target state
        if i is not None:
            df = pd.concat([df, i.run_data.sited_df])

    if write_output:

        # write output CSV
        out_csv = os.path.join(model.settings_dict.get('output_directory'), f"cerf_sited_{model.settings_dict.get('run_year')}_conus.csv")
        df.to_csv(out_csv, index=False)

    return df


def run(config_file=None, config_dict={}, write_output=True, n_jobs=-1, method='sequential',
            initialize_site_data=None, log_level='info'):
    """Run all CERF states for the CONUS for the target year.

    :param config_file:                 Full path with file name and extension to the input config.yml file
    :type config_file:                  str

    :param config_dict:                 Optional instead of config_file. Configuration dictionary.
    :type config_dict:                  dict

    :param write_output:                Write output as a raster to the output directory specified in the config file
    :type write_output:                 bool

    :param n_jobs:                      The number of processors to utilize.  Default is -1 which is all but 1.
    :type n_jobs:                       int

    :param method:                      Backend parallelization method used in Joblib.  Default is sequential to
                                        manage overhead for local runs.  Options for advanced configurations are:
                                        loky, threading, and multiprocessing.
                                        See https://joblib.readthedocs.io/en/latest/parallel.html for details.
    :type method:                       str

    :param initialize_site_data:        None if no initialization is required, otherwise either a CSV file or
                                        Pandas DataFrame of siting data bearing the following required fields:

                                        xcoord:  the X coordinate of the site in meters in
                                        USA_Contiguous_Albers_Equal_Area_Conic (EPSG:  102003)

                                        ycoord:  the Y coordinate of the site in meters in
                                        USA_Contiguous_Albers_Equal_Area_Conic (EPSG:  102003)

                                        retirement_year:  the year (int four digit, e.g., 2050) that the power
                                        plant is to be decommissioned

                                        buffer_in_km:  the buffer around the site to apply in kilometers

    :param log_level:                   Log level.  Options are 'info' and 'debug'.  Default 'info'
    :type log_level:                    str

    :return:                            A data frame containing each sited power plant and their attributes

    """

    try:

        # instantiate CERF model
        model = generate_model(config_file,
                               config_dict,
                               initialize_site_data=initialize_site_data,
                               log_level=log_level.lower())

        # process supporting data
        data = model.stage()

        # process all CERF CONUS states in parallel and store the result as a 2D arrays containing sites as
        #  the technology ID per grid cell.  All non-sited grid cells are given the value of NaN.
        df = cerf_parallel(model=model,
                           data=data,
                           write_output=write_output,
                           n_jobs=n_jobs,
                           method=method)

        logging.info(f"CERF model run completed in {round(time.time() - model.start_time, 7)} seconds")

    finally:
        # clean up logger
        model.close_logger()

    return df



import os
import pkg_resources

import yaml
import pandas as pd
import geopandas as gpd


def config_file(yr):
    """Return the sample configuration file for 2010.

    :param yr:                  Target four-digit year
    :type yr:                   int

    :return:                    Path to the target sample config file

    """

    return pkg_resources.resource_filename('cerf', f'data/config_{yr}.yml')


def cerf_states_shapefile():
    """Return the cerf CONUS states shapefile as a Geopandas data frame."""

    f = pkg_resources.resource_filename('cerf', 'data/cerf_conus_states_albers.zip')

    return gpd.read_file(f)


def cerf_boundary_shapefile():
    """Return the cerf CONUS boundary shapefile as a Geopandas data frame."""

    f = pkg_resources.resource_filename('cerf', 'data/cerf_conus_boundary_albers.zip')

    return gpd.read_file(f)


def cerf_crs():
    """Return a coordinate reference system (CRS) object of class 'pyproj.crs.crs.CRS'
     for USA_Contiguous_Albers_Equal_Area_Conic.

     """

    gdf = cerf_states_shapefile()

    return gdf.crs


def costs_per_kv_substation():
    """Return a dictionary of the cost of interconnection to substations of certain KV classes."""

    f = pkg_resources.resource_filename('cerf', 'data/costs_per_kv_substation.yml')

    with open(f, 'r') as yml:
        return yaml.load(yml, Loader=yaml.FullLoader)


def load_sample_config(yr):
    """Read the config YAML file for illustrative purposes.

    :param yr:                  Target configuration year in YYYY format.
    :type yr:                   int

    :return:                    dictionary for the configuration

    """

    available_years = list(range(2010, 2055, 5))

    if yr not in available_years:
        raise KeyError(f"Year '{yr}' not available as a default configuration file.  Must be in {available_years}")

    f = pkg_resources.resource_filename('cerf', f'data/config_{yr}.yml')

    with open(f, 'r') as yml:
        return yaml.load(yml, Loader=yaml.FullLoader)


def list_available_suitability_files():
    """Return a list of available suitability files."""

    root_dir = pkg_resources.resource_filename('cerf', 'data')

    return [os.path.join(root_dir, i) for i in os.listdir(root_dir) if
            (i.split('_')[0] == 'suitability') and
            (os.path.splitext(i)[-1] == '.sdat')]


def sample_lmp_zones_raster_file():
    """Return path for the sample lmp zoness raster file."""

    return pkg_resources.resource_filename('cerf', 'data/lmp_zones_1km.img')


def get_sample_lmp_data():
    """Return the sample 8760 hourly locational marginal price data as a Pandas DataFrame."""

    f = pkg_resources.resource_filename('cerf', 'data/illustrative_lmp_8760-per-zone_dollars-per-mwh.zip')

    return pd.read_csv(f)


def get_state_abbrev_to_name():
    """Return a dictionary of state abbreviation to state name."""

    # get state abbreviations file from cerf package data
    states_file = pkg_resources.resource_filename('cerf', 'data/state-abbrev_to_state-name.yml')

    # get state abbreviations to search for in HIFLD data
    with open(states_file, 'r') as yml:
        return yaml.load(yml, Loader=yaml.FullLoader)


def get_data_directory():
    """Return the directory of where the cerf package data resides."""

    return pkg_resources.resource_filename('cerf', 'data')



import logging

import numpy as np
import pandas as pd

import cerf.utils as util


class Competition:
    """Technology competition algorithm for CERF.

    Grid cell level net locational cost (NLC) per technology and an electricity technology capacity expansion plan are
    used to compete technologies against each other to see which will win the grid cell. The technology that wins the
    grid cell is then sited until no further winning cells exist. Once sited, the location of the winning technology’s
    grid cell, along with its buffer, are no longer available for siting. The competition array is recalculated after
    all technologies have passed through an iteration. This process is repeated until there are either no cells left
    to site or there are no more power plants left to satisfy the expansion plan for any technology. For technologies
    that have the same NLC value in multiple grid cells that win the competition, random selection is available by
    default. If the user wishes to have the outcomes be repeatable, the randomizer can be set to False and a random
    seed set.

    :param expansion_plan:                          Dictionary of {tech_id: number_of_sites, ...}
    :type expansion_plan:                           dict

    :param nlc_mask:                                3D masked array of [tech_id, x, y] for Net Locational Costs. Each
                                                    technology has been masked with its suitability data, so only
                                                    grid cells that are suitable have an NLC per tech. The 0 index
                                                    position is a default dimension which is chosen if no technologies
                                                    are able to compete.
    :type nlc_mask:                                 ndarray

    :param technology_dict:                         A technology dictionary containing at a minimum
                                                    {tech_id:  buffer_in_km, ...}
    :type technology_dict:                          dict

    :param randomize:                               Choose to make randomization of site selection where NLC is the same
                                                    in multiple grid cells for a single technology random. If False,
                                                    the seed_value will be used as a way to reproduce the exact siting.
                                                    Default:  True
    :type randomize:                                bool

    :param seed_value:                              Value for the see if randomize is False.
    :type seed_value:                               int

    :param verbose:                                 Log out siting information. Default False.
    :type verbose:                                  bool

    """

    def __init__(self,
                 target_state_name,
                 settings_dict,
                 technology_dict,
                 technology_order,
                 expansion_dict,
                 lmp_dict,
                 nov_dict,
                 ic_dict,
                 nlc_mask,
                 zones_arr,
                 xcoords,
                 ycoords,
                 indices_flat,
                 randomize=True,
                 seed_value=0,
                 verbose=False):

        # target state
        self.target_state_name = target_state_name

        # project level settings dictionary
        self.settings_dict = settings_dict

        # dictionary containing technology specific information
        self.technology_dict = technology_dict

        # order of technologies to process
        self.technology_order = technology_order

        # dictionary containing the expansion plan
        self.expansion_dict = expansion_dict

        # locational marginal pricing
        self.lmp_flat_dict = lmp_dict

        # net operational value
        self.nov_flat_dict = nov_dict

        # interconnection costs
        self.ic_flat_dict = ic_dict

        # lmp zones array
        self.zones_flat_arr = zones_arr

        # flat array of full grid indices value for the target state
        self.indices_flat = indices_flat

        # net locational costs with suitability mask for the target state
        self.nlc_mask = nlc_mask
        self.nlc_mask_shape = self.nlc_mask.shape

        # log out additional info
        self.verbose = verbose

        # use random seed to create reproducible outcomes
        if randomize is False:
            np.random.seed(seed_value)

        # number of technologies
        self.n_techs = len(self.technology_order)

        # dictionary to hold sited information
        self.sited_dict = util.empty_sited_dict()

        # coordinates for each index
        self.xcoords = xcoords
        self.ycoords = ycoords

        # show cheapest option, add 1 to the index to represent the technology number
        self.cheapest_arr = np.argmin(self.nlc_mask, axis=0)

        # flatten cheapest array to be able to use random
        self.cheapest_arr_1d = self.cheapest_arr.flatten()

        # prep array to hold outputs
        self.sited_arr_1d = np.zeros_like(self.cheapest_arr_1d)

        # set initial value to for available grid cells
        self.avail_grids = np.where(self.cheapest_arr_1d > 0)[0].shape[0]

        # create dictionary of {tech_id: flat_nlc_array, ...}
        self.nlc_flat_dict = {i: self.nlc_mask[ix+1, :, :].flatten() for ix, i in enumerate(self.technology_order)}

        # run competition and site
        self.sited_array, self.sited_df = self.compete()

        # evaluate sites to see if expansion plan was met
        self.log_outcome()

    def log_outcome(self):
        """Log a warning sites that were not able to be sited."""

        for k in self.expansion_dict.keys():

            tech_name = self.expansion_dict[k]['tech_name']
            remaining_sites = self.expansion_dict[k]['n_sites']

            if remaining_sites > 0:
                logging.warning(f"Unable to achieve full siting for `{tech_name}` in `{self.target_state_name}`:  {remaining_sites} unsited.")

    def compete(self):

        # initialize keep sighting designation; False if no more sites or area to site
        keep_siting = True

        while keep_siting:

            # evaluate by technology
            for index, tech_id in enumerate(self.technology_order):

                # assign an index as it appears in the n-dim array to the order in which it is being processed
                #  index of 0 is the default array and does not represent a technology
                tech_index = index + 1

                # get the indices of the target tech ids where the target tech is the cheapest option
                tech = np.where(self.cheapest_arr_1d == tech_index)[0]

                # the number of sites for the target tech
                required_sites = self.expansion_dict[tech_id]['n_sites']

                # calculate the year of retirement
                retirement_year = self.settings_dict['run_year'] + int(self.technology_dict[tech_id]['lifetime'])

                # if there are more power plants to site and there are grids available to site them...
                if self.avail_grids > 0 and tech.shape[0] > 0 and required_sites > 0:

                    # site with buffer and exclude buffered area from further siting
                    still_siting = True
                    sited_list = []
                    while still_siting:

                        # get the NLC values associated with each winner
                        tech_nlc = self.nlc_flat_dict[tech_id][tech]

                        # get the least expensive NLC indices from the winners
                        tech_nlc_cheap = tech[np.where(tech_nlc == np.nanmin(tech_nlc))]

                        # select a random index that has a winning cell for the check where multiple low NLC may exists
                        target_ix = np.random.choice(tech_nlc_cheap)

                        # add selected index to sited dictionary
                        self.sited_dict['state_name'].append(self.target_state_name)
                        self.sited_dict['tech_id'].append(tech_id)
                        self.sited_dict['tech_name'].append(self.technology_dict[tech_id]['tech_name'])
                        self.sited_dict['xcoord'].append(self.xcoords[target_ix])
                        self.sited_dict['ycoord'].append(self.ycoords[target_ix])
                        self.sited_dict['index'].append(self.indices_flat[target_ix])
                        self.sited_dict['buffer_in_km'].append(self.technology_dict[tech_id]['buffer_in_km'])
                        self.sited_dict['sited_year'].append(self.settings_dict['run_year'])
                        self.sited_dict['retirement_year'].append(retirement_year)
                        self.sited_dict['lmp_zone'].append(self.zones_flat_arr[target_ix])
                        self.sited_dict['locational_marginal_pricing'].append(self.lmp_flat_dict[tech_id][target_ix])
                        self.sited_dict['net_operational_value'].append(self.nov_flat_dict[tech_id][target_ix])
                        self.sited_dict['interconnection_cost'].append(self.ic_flat_dict[tech_id][target_ix])
                        self.sited_dict['net_locational_cost'].append(self.nlc_flat_dict[tech_id][target_ix])

                        # add selected index to list
                        sited_list.append(target_ix)

                        # apply buffer
                        result = util.buffer_flat_array(target_index=target_ix,
                                                   arr=self.cheapest_arr_1d,
                                                   nrows=self.cheapest_arr.shape[0],
                                                   ncols=self.cheapest_arr.shape[1],
                                                   ncells=self.technology_dict[tech_id]['buffer_in_km'],
                                                   set_value=0)

                        # unpack values
                        self.cheapest_arr_1d, buffer_indices_list = result

                        # update the number of sites left to site
                        required_sites -= 1
                        self.expansion_dict[tech_id].update(n_sites=required_sites)

                        # remove any buffered elements as an option to site
                        tech_indices_to_delete = [np.where(tech == i)[0][0] for i in buffer_indices_list if i in tech]
                        tech = np.delete(tech, tech_indices_to_delete)

                        # exit siting for the target technology if all sites have been sited or if there are no more
                        #   winning cells
                        if required_sites == 0 or tech.shape[0] == 0:
                            still_siting = False

                    # array of the site indices
                    rdx = np.array(sited_list)

                    # add sited techs to output array
                    self.sited_arr_1d[rdx] = tech_id

                    if self.verbose:
                        logging.info('\nUpdate expansion plan to represent siting requirements:')
                        logging.info(self.expansion_dict)

                    # update original array with excluded area where siting occurred
                    # if target technology has no more sites to be sited
                    if self.expansion_dict[tech_id] == 0:

                        # make all elements for the target tech in the NLC mask unsuitable so we can progress
                        self.nlc_mask[tech_index, :, :] = np.ma.masked_array(self.nlc_mask[0, :, :],
                                                                             np.ones_like(self.nlc_mask[0, :, :]))

                    # apply the new exclusion from the current technology to all techs...
                    #   invert sited elements to have a value of 1 so they can be used as a mask
                    #   repeat the new sited array to create a mask for all techs and reshape to 2D
                    #   update all technologies with the new mask
                    self.nlc_mask[1:, :, :] = np.ma.masked_array(self.nlc_mask[1:, :, :],
                                                                 np.tile(np.where(self.cheapest_arr_1d == 0, 1, 0),
                                                                         self.nlc_mask_shape[0] - 1).reshape(
                                                                     (self.nlc_mask_shape[0] - 1,
                                                                      self.nlc_mask_shape[1],
                                                                      self.nlc_mask_shape[2])))

                    # if the technology has achieved its full expansion, then mask the rest of its suitable area so
                    #  other technologies can now compete for the grid cells it previously won but now no longer needs
                    if self.expansion_dict[tech_id]['n_sites'] == 0:
                        self.nlc_mask[tech_index, :, :] = np.ma.masked_array(self.nlc_mask[tech_index, :, :],
                                                                             np.ones_like(self.nlc_mask[tech_index, :, :]))

                    # show cheapest option, add 1 to the index to represent the technology number
                    self.cheapest_arr = np.argmin(self.nlc_mask, axis=0)

                    # flatten cheapest array to be able to use random
                    self.cheapest_arr_1d = self.cheapest_arr.flatten()

                    # check for any available grids to site in
                    self.avail_grids = np.where(self.cheapest_arr_1d > 0)[0].shape[0]

                    # are there any sites left to site
                    left_to_site = sum([self.expansion_dict[i]['n_sites'] for i in self.expansion_dict.keys()])

                    # stop technology iteration if all area is consumed or if all sites have been sited
                    if self.avail_grids == 0 or left_to_site == 0:
                        keep_siting = False

                    if self.verbose:
                        logging.info(f'\nAvailable grid cells:  {self.avail_grids}')

                # there are no more suitable grid cells
                elif self.avail_grids == 0:
                    keep_siting = False

                # if there are available grids but n
                elif self.avail_grids > 0 and tech.shape[0] > 0 and required_sites == 0:

                    # if there are no required sites, then mask the rest of the techs suitable area so
                    #  other technologies can now compete for the grid cells it previously won but now no longer needs
                    self.nlc_mask[tech_index, :, :] = np.ma.masked_array(self.nlc_mask[tech_index, :, :],
                                                                         np.ones_like(self.nlc_mask[tech_index, :, :]))

                    # show cheapest option, add 1 to the index to represent the technology number
                    self.cheapest_arr = np.argmin(self.nlc_mask, axis=0)

                    # flatten cheapest array to be able to use random
                    self.cheapest_arr_1d = self.cheapest_arr.flatten()

                    # check for any available grids to site in
                    self.avail_grids = np.where(self.cheapest_arr_1d > 0)[0].shape[0]

                # create sited data frame
                df = pd.DataFrame(self.sited_dict).astype(util.sited_dtypes())

        # reshape output array to 2D
        return self.sited_arr_1d.reshape(self.cheapest_arr.shape), df



"""Logger for CERF model.

Copyright (c) 2018, Battelle Memorial Institute

Open source under license BSD 2-Clause - see LICENSE and DISCLAIMER

@author:  Chris R. Vernon (chris.vernon@pnnl.gov)

"""

import logging
import sys


class Logger:
    """Initialize project-wide logger. The logger outputs to both stdout and a file."""

    # output format for log string
    LOG_FORMAT_STRING = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    @property
    def log_format(self):
        """Generate log formatter."""

        return logging.Formatter(self.LOG_FORMAT_STRING)

    @property
    def logger(self):
        """Initialize logger as level info."""

        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

        return logger

    def initialize_logger(self):
        """Initialize logger to stdout and file."""

        # logger console handler
        self.console_handler()

        # logger file handler
        if self.write_logfile:
            self.file_handler()

    def console_handler(self, log_level):
        """Construct console handler."""

        console_handler = logging.StreamHandler(sys.stdout)

        if log_level == 'debug':
            console_handler.setLevel(logging.DEBUG)
        else:
            console_handler.setLevel(logging.INFO)

        console_handler.setFormatter(self.log_format)
        self.logger.addHandler(console_handler)

    def file_handler(self, log_level):
        """Construct file handler."""

        file_handler = logging.FileHandler(self.logfile)

        if log_level == 'debug':
            file_handler.setLevel(logging.DEBUG)
        else:
            file_handler.setLevel(logging.INFO)

        file_handler.setFormatter(self.log_format)
        self.logger.addHandler(file_handler)

    @staticmethod
    def close_logger():
        """Shutdown logger."""

        # Remove logging handlers
        logger = logging.getLogger()

        for handler in logger.handlers[:]:
            handler.close()
            logger.removeHandler(handler)

        logging.shutdown()



"""Tests for spatial_reader.py

@author Chris R. vernon
@email chris.vernon@pnnl.gov

License:  BSD 2-Clause, see LICENSE and DISCLAIMER files

"""
import unittest

#import cerf.spatial_reader as sr


class TestSpatialReader(unittest.TestCase):
    """Tests for `spatial_reader.py`"""

    # BIOMASS_DIR = pkg_resources.resource_dir('cerf', 'tests/data/inputs/spatial/biomass')
    # NUCLEAR_DIR = pkg_resources.resource_dir('cerf', 'tests/data/inputs/spatial/nuclear')

    def test_sum_rasters(self):
        """Tests for `sum_rasters` function."""

        # arr = sr.sum_rasters(TestSpatialReader.BIOMASS_DIR)
        #
        # print(arr.shape)
        # print(arr.min())
        # print(arr.max())
        # print(arr.unique())

        self.assertEqual(2, 2)
        pass


if __name__ == '__main__':
    unittest.main()



import unittest
import pkg_resources

import pandas as pd
import geopandas as gpd
import yaml

from cerf.package_data import *


class TestPackageData(unittest.TestCase):
    """Tests for package data matching to confirm load function modification does not happen."""

    def test_config_file(self):
        """Ensure package data functions do not get modified."""

        yr = 2010

        comp = pkg_resources.resource_filename('cerf', f'data/config_{yr}.yml')
        val = config_file(yr)

        self.assertEqual(comp, val)

    def test_cerf_states_shapefile(self):
        """Ensure package data functions do not get modified."""

        comp = gpd.read_file(pkg_resources.resource_filename('cerf', 'data/cerf_conus_states_albers.zip'))
        val = cerf_states_shapefile()

        pd.testing.assert_frame_equal(comp, val)

    def test_cerf_boundary_shapefile(self):
        """Ensure package data functions do not get modified."""

        comp = gpd.read_file(pkg_resources.resource_filename('cerf', 'data/cerf_conus_boundary_albers.zip'))
        val = cerf_boundary_shapefile()

        pd.testing.assert_frame_equal(comp, val)

    def test_cerf_crs(self):
        """Ensure package data functions do not get modified."""

        comp = cerf_states_shapefile().crs
        val = cerf_crs()

        self.assertEqual(comp, val)

    def test_costs_per_kv_substation(self):
        """Ensure package data functions do not get modified."""

        f = pkg_resources.resource_filename('cerf', 'data/costs_per_kv_substation.yml')

        with open(f, 'r') as yml:
            comp = yaml.load(yml, Loader=yaml.FullLoader)

        val = costs_per_kv_substation()

        self.assertEqual(comp, val)

    def test_load_sample_config(self):
        """Ensure package data functions do not get modified."""

        yr = 2010

        available_years = list(range(2010, 2055, 5))

        if yr not in available_years:
            raise KeyError(f"Year '{yr}' not available as a default configuration file.  Must be in {available_years}")

        f = pkg_resources.resource_filename('cerf', f'data/config_{yr}.yml')

        with open(f, 'r') as yml:
            comp =  yaml.load(yml, Loader=yaml.FullLoader)

        val = load_sample_config(yr)

        self.assertEqual(comp, val)

    def test_list_available_suitability_files(self):
        """Ensure package data functions do not get modified."""

        root_dir = pkg_resources.resource_filename('cerf', 'data')

        comp = [os.path.join(root_dir, i) for i in os.listdir(root_dir) if
                (i.split('_')[0] == 'suitability') and
                (os.path.splitext(i)[-1] == '.sdat')]

        val = list_available_suitability_files()

        self.assertEqual(comp, val)

    def test_sample_lmp_zones_raster_file(self):
        """Ensure package data functions do not get modified."""

        comp = pkg_resources.resource_filename('cerf', 'data/lmp_zones_1km.img')
        val = sample_lmp_zones_raster_file()

        self.assertEqual(comp, val)

    def test_get_sample_lmp_data(self):
        """Ensure package data functions do not get modified."""

        comp = pd.read_csv(pkg_resources.resource_filename('cerf', 'data/illustrative_lmp_8760-per-zone_dollars-per-mwh.zip'))
        val = get_sample_lmp_data()

        pd.testing.assert_frame_equal(comp, val)

    def test_get_state_abbrev_to_name(self):
        """Ensure package data functions do not get modified."""

        states_file = pkg_resources.resource_filename('cerf', 'data/state-abbrev_to_state-name.yml')

        with open(states_file, 'r') as yml:
            comp = yaml.load(yml, Loader=yaml.FullLoader)

        val = get_state_abbrev_to_name()

        self.assertEqual(comp, val)

    def test_get_data_directory(self):
        """Ensure package data functions do not get modified."""

        comp = pkg_resources.resource_filename('cerf', 'data')
        val = get_data_directory()

        self.assertEqual(comp, val)


if __name__ == '__main__':
    unittest.main()



"""Tests to ensure high-level functionality and outputs remain consistent.

@author Chris R. vernon
@email chris.vernon@pnnl.gov

License:  BSD 2-Clause, see LICENSE and DISCLAIMER files

"""
import unittest


class TestReadConfig(unittest.TestCase):
    """Test configuration reader."""

    # CONFIG_FILE = pkg_resources.resource_filename('cerf', 'tests/data/config.yml')

    def test_config_reader(self):
        """Tests for configuration reader."""

        self.assertEqual(2, 2)
        pass


if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd

import logging

import pkg_resources


def generate_random_lmp_dataframe(n_zones=57, low_value=10, mid_value=300, high_value=500, n_samples=5000):
    """Generate a random dataframe of hourly 8760 LMP values per lmp zone.  Let high value LMPs only be used
    for 15 percent of the data.
    :param n_zones:                     Number of zones to process
    :param low_value:                   Desired minimum value of MWh
    :param mid_value:                   Desired mid value of MWh to split the 85-15 split to
    :param high_value:                  Desired max value of MWh
    :param n_samples:                   Number of intervals to split the min, max choices by
    :return:                            Data frame of LMPs per zone
    """

    # initialize a dictionary with the hour count for the number of hours in a year
    d = {'hour': list(range(1, 8761, 1))}

    # create an array with n_samples covering an equal space from low to mid values
    array_1 = np.linspace(low_value, mid_value, n_samples)

    # create an array with n_samples covering an equal space from mid to low values
    array_2 = np.linspace(mid_value, high_value, n_samples)

    # let only 15 percent of values come from high cost values
    threshold = 8760 - (8760 * 0.15)

    # create an LMP array for each zone
    for i in range(n_zones):

        # construct a list of random LMP values
        l = []
        for j in range(8760):

            if j < threshold:
                l.append(np.random.choice(array_1))

            else:
                l.append(np.random.choice(array_2))

        # shuffle the list
        np.random.shuffle(l)

        # assign to dict
        d[i] = l

    # convert to data frame
    return pd.DataFrame(d)


class LocationalMarginalPricing:
    """Create a 3D array of locational marginal pricing per technology by capacity factor.

    Locational Marginal Pricing (LMP) represents the cost of making and delivering electricity
    over an interconnected network of service nodes. LMPs are delivered on an hourly basis
    (8760 hours for the year) and help us to understand aspects of generation and congestion
    costs relative to the supply and demand of electricity when considering existing transmission
    infrastructure.  LMPs are a also driven by factors such as the cost of fuel which cerf also
    takes into account when calculating a power plants :ref:`Net Operating Value`.  When working
    with a scenario-driven grid operations model to evaluate the future evolution of the electricity
    system, **cerf** can ingest LMPs, return the sited generation per service area for the time
    step, and then continue this iteration through all future years to provide a harmonized view
    how the electricity system may respond to stressors in the future.

    :param lmp_zone_dict:                      A dictionary containing lmp related settings from the config file
    :type lmp_zone_dict:                       dict

    :param technology_dict:                    A dictionary containing technology related settings from the config file
    :type technology_dict:                     dict

    :param technology_order:                   A list of technologies in the order by which they should be processed
    :type lmp_zone_dict:                       list

    :param zones_arr:                          An array containing the lmp zones per grid cell
    :type lmp_zone_dict:                       dict

    """

    def __init__(self, lmp_zone_dict, technology_dict, technology_order, zones_arr):

        # dictionary containing lmp zones information
        self.lmp_zone_dict = lmp_zone_dict

        # dictionary containing technology specific information
        self.technology_dict = technology_dict

        # order of technologies to process
        self.technology_order = technology_order

        # array containing the lmp zones per grid cell
        self.zones_arr = zones_arr

    @staticmethod
    def get_cf_bin(capacity_factor):
        """Get the correct start and through index values to average over for calculating LMP."""

        if capacity_factor == 1.0:
            start_index = 0
            through_index = 8760

        elif capacity_factor >= 0.5:
            start_index = int(np.ceil(8760 * (1 - capacity_factor)))
            through_index = 8760

        elif capacity_factor == 0.0:
            msg = f"The capacity factor provided `{capacity_factor}` is outside the bounds of 0.0 through 1.0"
            raise ValueError(msg)

        else:
            start_index = 0
            through_index = int(np.ceil(8760 * capacity_factor))

        return start_index, through_index

    def get_lmp(self):
        """Create LMP array for the current technology.

        :return:                    3D numpy array of LMP where [tech_id, x, y]

        """

        # number of technologies
        n_technologies = len(self.technology_dict)

        lmp_arr = np.zeros(shape=(n_technologies, self.zones_arr.shape[0], self.zones_arr.shape[1]))

        # get the LMP file for the technology from the configuration file
        lmp_file = self.lmp_zone_dict.get('lmp_hourly_data_file', None)

        # use illustrative default if none provided
        if lmp_file is None:

            # default illustrative LMP file
            lmp_file = pkg_resources.resource_filename('cerf', 'data/illustrative_lmp_8760-per-zone_dollars-per-mwh.zip')
            logging.info(f"Using LMP from default illustrative package data:  {lmp_file}")

        else:
            logging.info(f"Using LMP file:  {lmp_file}")

        lmp_df = pd.read_csv(lmp_file)

        # drop the hour field
        lmp_df.drop('hour', axis=1, inplace=True)

        for index, i in enumerate(self.technology_order):

            # assign the correct LMP based on the capacity factor of the technology
            start_index, through_index = self.get_cf_bin(self.technology_dict[i]['capacity_factor'])

            # sort by descending lmp for each zone
            for j in lmp_df.columns:
                lmp_df[j] = lmp_df[j].sort_values(ascending=False).values

            # create a dictionary of LMP values for each power zone based on tech capacity factor
            lmp_dict = lmp_df.iloc[start_index:through_index].mean(axis=0).to_dict()
            lmp_dict = {int(k): lmp_dict[k] for k in lmp_dict.keys()}

            # add in no data
            lmp_dict[self.lmp_zone_dict['lmp_zone_raster_nodata_value']] = np.nan

            # create LMP array for the current technology
            lmp_arr[index, :, :] = np.vectorize(lmp_dict.get)(self.zones_arr)

        return lmp_arr



"""Tests for CERF utilities.

:author:   Chris R. Vernon
:email:    chris.vernon@pnnl.gov

License:  BSD 2-Clause, see LICENSE and DISCLAIMER files

"""

import unittest

import numpy as np

import cerf.utils as util


class TestUtils(unittest.TestCase):
    """Tests for the NOV calculations."""

    # comparison data
    COMP_BUFF_FLAT_0 = np.array([[7., 7., 7., 0., 0.],
                                 [7., 7., 7., 0., 0.],
                                 [7., 7., 7., 0., 0.],
                                 [0., 0., 0., 0., 0.]])

    COMP_BUFF_FLAT_0_LIST = [0, 1, 2, 5, 6, 7, 10, 11, 12]

    COMP_BUFF_FLAT_19 = np.array([[0., 0., 0., 0., 0.],
                                  [0., 0., 7., 7., 7.],
                                  [0., 0., 7., 7., 7.],
                                  [0., 0., 7., 7., 7.]])

    COMP_BUFF_FLAT_19_LIST = [17, 18, 19, 12, 13, 14, 7, 8, 9]

    def test_buffer_flat_array(self):
        """Test to make sure the buffer flat array function if working correctly."""

        # create sample array
        arr = np.zeros(shape=(4, 5))

        # flatten array
        arr_1d = arr.flatten()

        # buffer top left corner by two cells and set the value to 7
        arr_1d_0, buff_0 = util.buffer_flat_array(target_index=0,
                                                  arr=arr_1d.copy(),
                                                  nrows=arr.shape[0],
                                                  ncols=arr.shape[1],
                                                  ncells=2,
                                                  set_value=7)

        # reshape to 2D
        arr_2d_0 = arr_1d_0.reshape(arr.shape)

        # compare array
        np.testing.assert_array_equal(TestUtils.COMP_BUFF_FLAT_0, arr_2d_0)

        # compare buffer indices
        self.assertEqual(TestUtils.COMP_BUFF_FLAT_0_LIST, buff_0)

        # buffer bottom right corner by two cells and set the value to 7
        arr_1d_19, buff_19 = util.buffer_flat_array(target_index=19,
                                                    arr=arr_1d.copy(),
                                                    nrows=arr.shape[0],
                                                    ncols=arr.shape[1],
                                                    ncells=2,
                                                    set_value=7)

        # reshape to 2D
        arr_2d_19 = arr_1d_19.reshape(arr.shape)

        # compare array
        np.testing.assert_array_equal(TestUtils.COMP_BUFF_FLAT_19, arr_2d_19)

        # compare buffer indices
        self.assertEqual(TestUtils.COMP_BUFF_FLAT_19_LIST, buff_19)


if __name__ == '__main__':
    unittest.main()


import unittest

import numpy as np

from cerf.compete import Competition


class TestCompete(unittest.TestCase):

    EXPANSION_PLAN = {1: {'n_sites': 1, 'tech_name': 'test1'},
                      2: {'n_sites': 1, 'tech_name': 'test2'},
                      3: {'n_sites': 1, 'tech_name': 'test3'}}  # n sites per tech
    TECH_DICT = {1: {'buffer_in_km': 1, 'lifetime': 60, 'tech_name': 'test1'},
                 2: {'buffer_in_km': 1, 'lifetime': 60, 'tech_name': 'test2'},
                 3: {'buffer_in_km': 1, 'lifetime': 60, 'tech_name': 'test3'}}  # buffer per tech
    TECH_ORDER = [1, 2, 3]

    # proxy NLC array
    NLC_ARR = np.array([[[1.2, 3.2, 3, 3.2, 3, 3.2, 3],
                         [2.4, 5, 2.4, 3.2, 3, 3.2, 3],
                         [7, 2.4, 9, 3.2, 3, 3.2, 3],
                         [2.4, 5, 2.4, 3.2, 3, 3.2, 3]],
                        [[2, 1, 1.4, 3.2, 3, 3.2, 3],
                         [5, 5, 7, 3.2, 3, 3.2, 3],
                         [1, 9, 3, 3.2, 3, 3.2, 3],
                         [2.4, 5, 2.4, 3.2, 3, 3.2, 3]],
                        [[0, 1, 0, 3.2, 3, 3.2, 3],
                         [5, 4, 7, 3.2, 3, 3.2, 3],
                         [1, 9, 9, 3.2, 3, 3.2, 3],
                         [2.4, 5, 2.4, 3.2, 3, 3.2, 3]]])

    # excluded by suitability 1=not suitable, 0=suitable
    SUIT_ARR = np.array([[[0, 0, 1, 0, 1, 0, 1], [1, 0, 1, 0, 1, 0, 1], [0, 0, 1, 0, 1, 0, 1], [1, 0, 1, 0, 1, 0, 1]],
                        [[1, 0, 1, 0, 1, 0, 1], [1, 0, 1, 0, 1, 0, 1], [1, 0, 0, 0, 1, 0, 1], [1, 0, 1, 0, 1, 0, 1]],
                        [[0, 0, 0, 0, 1, 0, 1], [0, 0, 0, 0, 1, 0, 1], [1, 1, 1, 0, 1, 0, 1], [1, 0, 1, 0, 1, 0, 1]]])

    # trimmed down settings dictionary
    SETTINGS_DICT = {'run_year': 2010}

    # expected outcome
    COMP_SITED = np.array([[0, 2, 0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0, 0, 0],
                           [0, 1, 0, 3, 0, 0, 0],
                           [0, 0, 0, 0, 0, 0, 0]])

    COMP_EXP_PLAN = {1: {'n_sites': 0, 'tech_name': 'test1'},
                     2: {'n_sites': 0, 'tech_name': 'test2'},
                     3: {'n_sites': 0, 'tech_name': 'test3'}}

    COMP_SITED_DICT = {'state_name': ['test', 'test', 'test'],
                       'index': [2.4, 3.2, 3.2],
                       'tech_id': [1, 2, 3],
                       'tech_name': ['test1', 'test2', 'test3'],
                       'xcoord': [2.4, 3.2, 3.2],
                       'ycoord': [2.4, 3.2, 3.2],
                       'buffer_in_km': [1, 1, 1],
                       'sited_year': [2010, 2010, 2010],
                       'retirement_year': [2070, 2070, 2070],
                       'lmp_zone': [2, 3, 3],
                       'locational_marginal_pricing': [2.4, 1.0, 3.2],
                       'net_operational_value': [2.4, 1.0, 3.2],
                       'interconnection_cost': [2.4, 1.0, 3.2],
                       'net_locational_cost': [2.4, 1.0, 3.2]}

    @classmethod
    def create_masked_nlc_array(cls):
        """Create a masked NLC array by suitability to use in testing."""

        # insert zero array and mask it as index [0, :, :] so the tech_id 0 will always be min if nothing is left to site
        arr = np.insert(cls.NLC_ARR, 0, np.zeros_like(cls.NLC_ARR[0, :, :]), axis=0)

        # exclude all area for the proxy dimension
        exc = np.insert(cls.SUIT_ARR, 0, np.ones_like(cls.SUIT_ARR[0, :, :]), axis=0)

        # apply exclusion
        return np.ma.masked_array(arr, exc)

    @classmethod
    def create_proxy_arrays(cls):
        """Create fake arrays to feed into the class."""

        # fake dictionary
        fake_dict = {i + 1: cls.NLC_ARR[i, :, :].flatten() for i in range(cls.NLC_ARR.shape[0])}

        # fake flat array
        fake_flat_arr = fake_dict[1].flatten()

        return fake_dict, fake_flat_arr

    def test_competition(self):
        """Ensure that the competition algorithm performs as expected."""

        # create a fake NLC masked array to use for testing
        nlc_arr = self.create_masked_nlc_array()

        # proxy dictionary and arrays
        fake_dict, fake_flat_array = self.create_proxy_arrays()

        comp = Competition(target_state_name='test',
                           settings_dict=TestCompete.SETTINGS_DICT,
                           technology_dict=TestCompete.TECH_DICT,
                           technology_order=TestCompete.TECH_ORDER,
                           expansion_dict=TestCompete.EXPANSION_PLAN,
                           lmp_dict=fake_dict,
                           nov_dict=fake_dict,
                           ic_dict=fake_dict,
                           nlc_mask=nlc_arr,
                           zones_arr=fake_flat_array.astype(np.int32),
                           xcoords=fake_flat_array,
                           ycoords=fake_flat_array,
                           indices_flat=fake_flat_array,
                           randomize=False,
                           seed_value=0,
                           verbose=False)

        # test output equality
        np.testing.assert_array_equal(TestCompete.COMP_SITED, comp.sited_array)

        # ensure the expansion plan was updated
        self.assertEqual(TestCompete.COMP_EXP_PLAN, comp.expansion_dict)

        # check sited dict match
        self.assertEqual(TestCompete.COMP_SITED_DICT, comp.sited_dict)


if __name__ == '__main__':
    unittest.main()



"""Tests for Net Operational Value calculations.

:author:   Chris R. Vernon
:email:    chris.vernon@pnnl.gov

License:  BSD 2-Clause, see LICENSE and DISCLAIMER files

"""

import unittest

import numpy as np

from cerf.nov import NetOperationalValue


class TestNov(unittest.TestCase):
    """Tests for the NOV calculations."""

    # inputs as they would come from GCAM and input sources
    DISCOUNT_RATE = 0.05  # fraction
    LIFETIME = 60  # years
    UNIT_SIZE = 1350  # megawatt
    CAPACITY_FACTOR = 0.9  # fraction
    VARIABLE_COST_ESC_RATE = -0.00104311614063357  # fraction
    FUEL_ESC_RATE = 0.04639  # fraction
    CARBON_ESC_RATE = 0.0  # fraction
    VARIABLE_OM = 2.09812782440284  # $/MWh
    HEAT_RATE = 10246.19999999998  # Btu/kWh
    FUEL_PRICE = 0.712809999999999  # $/GJ gets converted to $/MBtu in code
    CARBON_TAX = 0.0  # $/ton
    CARBON_CAPTURE_RATE = 0.0  # fraction
    FUEL_CO2_CONTENT = 0.0  # tons/MWh gets converted to tons/Btu in code
    LMP_ARR = np.array([66.95609874])  # $/MWh

    # expected values
    EXPECTED_ANNUITY_FACTOR = 0.05282818452724236
    EXPECTED_LF_VOM = 0.9819019248539469
    EXPECTED_LF_FUEL = 2.8587075325530398
    EXPECTED_LF_CARBON = 0.999999999999999

    # difference in expected values when evaluating under different conditions
    EXPECTED_NOV_NOCARBON_NOLEAP = np.array([1780847345.1014912])
    EXPECTED_NOV_NOCARBON_LEAP = np.array([1785726378.923687])
    EXPECTED_NOV_WITHCARBON_NOLEAP = np.array([1780847344.7371392])

    @classmethod
    def instantiate_nov(cls, target_year, carbon_tax, fuel_co2_content, carbon_capture_rate, consider_leap_year):
        """Instantiate NOV class with test values.  The additional parameters can be passed to
        test NOV under different carbon conditions."""

        return NetOperationalValue(discount_rate=cls.DISCOUNT_RATE,
                                   lifetime=cls.LIFETIME,
                                   unit_size=cls.UNIT_SIZE,
                                   capacity_factor=cls.CAPACITY_FACTOR,
                                   variable_cost_esc_rate=cls.VARIABLE_COST_ESC_RATE,
                                   fuel_esc_rate=cls.FUEL_ESC_RATE,
                                   carbon_esc_rate=cls.CARBON_ESC_RATE,
                                   variable_om=cls.VARIABLE_OM,
                                   heat_rate=cls.HEAT_RATE,
                                   fuel_price=cls.FUEL_PRICE,
                                   carbon_tax=carbon_tax,
                                   carbon_capture_rate=carbon_capture_rate,
                                   fuel_co2_content=fuel_co2_content,
                                   lmp_arr=cls.LMP_ARR,
                                   target_year=target_year,
                                   consider_leap_year=consider_leap_year)

    def test_nocarbon_noleap_nov(self):
        """Test NOV outcome with no carbon and no leap year."""

        # create a no carbon, no leap year run
        econ = self.instantiate_nov(target_year=2010,  # four digit year
                                    carbon_tax=0.0,  # $/ton
                                    fuel_co2_content=0.0,  # tons/MWh gets converted to tons/Btu
                                    carbon_capture_rate=0.0,  # fraction
                                    consider_leap_year=False
                                    )
        nov_tech_arr = econ.calc_nov()

        # test the calculation of annuity factor
        self.assertEqual(TestNov.EXPECTED_ANNUITY_FACTOR, econ.annuity_factor)

        # test the calculation of the levelization factor for variable OM
        self.assertEqual(TestNov.EXPECTED_LF_VOM, econ.lf_vom)

        # test the calculation of the levelization factor for fuel
        self.assertEqual(TestNov.EXPECTED_LF_FUEL, econ.lf_fuel)

        # test the calculation of the levelization factor for carbon
        self.assertEqual(TestNov.EXPECTED_LF_CARBON, econ.lf_carbon)

        # test NOV
        np.testing.assert_array_equal(TestNov.EXPECTED_NOV_NOCARBON_NOLEAP, nov_tech_arr)

    def test_nocarbon_leap_nov(self):
        """Test NOV outcome with no carbon and on a leap year."""

        # create a no carbon, leap year run
        econ = self.instantiate_nov(target_year=2012,  # four digit year
                                    carbon_tax=0.0,  # $/ton
                                    fuel_co2_content=0.0,  # tons/MWh gets converted to tons/Btu
                                    carbon_capture_rate=0.0,  # fraction
                                    consider_leap_year=True
                                    )
        nov_tech_arr = econ.calc_nov()

        # test the calculation of annuity factor
        self.assertEqual(TestNov.EXPECTED_ANNUITY_FACTOR, econ.annuity_factor)

        # test the calculation of the levelization factor for variable OM
        self.assertEqual(TestNov.EXPECTED_LF_VOM, econ.lf_vom)

        # test the calculation of the levelization factor for fuel
        self.assertEqual(TestNov.EXPECTED_LF_FUEL, econ.lf_fuel)

        # test the calculation of the levelization factor for carbon
        self.assertEqual(TestNov.EXPECTED_LF_CARBON, econ.lf_carbon)

        # test NOV
        np.testing.assert_array_equal(TestNov.EXPECTED_NOV_NOCARBON_LEAP, nov_tech_arr)

    def test_with_carbon_noleap_nov(self):
        """Test NOV outcome with carbon and not on a leap year."""

        # create a no carbon, leap year run
        econ = self.instantiate_nov(target_year=2010,  # four digit year
                                    carbon_tax=10.0,  # $/ton
                                    fuel_co2_content=1.2,  # tons/MWh gets converted to tons/Btu
                                    carbon_capture_rate=0.05,  # fraction
                                    consider_leap_year=False
                                    )
        nov_tech_arr = econ.calc_nov()

        # test the calculation of annuity factor
        self.assertEqual(TestNov.EXPECTED_ANNUITY_FACTOR, econ.annuity_factor)

        # test the calculation of the levelization factor for variable OM
        self.assertEqual(TestNov.EXPECTED_LF_VOM, econ.lf_vom)

        # test the calculation of the levelization factor for fuel
        self.assertEqual(TestNov.EXPECTED_LF_FUEL, econ.lf_fuel)

        # test the calculation of the levelization factor for carbon
        self.assertEqual(TestNov.EXPECTED_LF_CARBON, econ.lf_carbon)

        # test NOV
        np.testing.assert_array_equal(TestNov.EXPECTED_NOV_WITHCARBON_NOLEAP, nov_tech_arr)


if __name__ == '__main__':
    unittest.main()



"""Stage data for CERF run.

@author Chris R. vernon
@email chris.vernon@pnnl.gov

License:  BSD 2-Clause, see LICENSE and DISCLAIMER files

"""

import logging

import numpy as np
import pkg_resources
import rasterio

import cerf.utils as util
from cerf.lmp import LocationalMarginalPricing
from cerf.nov import NetOperationalValue
from cerf.interconnect import Interconnection


class Stage:

    # type hints
    settings_dict: dict
    lmp_zone_dict: dict
    technology_dict: dict
    technology_order: list

    # default cerf technology name to suitability file dictionary
    DEFAULT_SUITABILITY = {'biomass_conv_wo_ccs': 'suitability_biomass.sdat',
                             'biomass_conv_w_ccs': 'suitability_biomass.sdat',
                             'biomass_igcc_wo_ccs': 'suitability_biomass_igcc.sdat',
                             'biomass_igcc_w_ccs': 'suitability_biomass_igcc_ccs.sdat',
                             'coal_conv_pul_wo_ccs': 'suitability_coal.sdat',
                             'coal_conv_pul_w_ccs': 'suitability_coal.sdat',
                             'coal_igcc_wo_ccs': 'suitability_coal_igcc.sdat',
                             'coal_igcc_w_ccs': 'suitability_coal_igcc_ccs.sdat',
                             'gas_cc_wo_ccs': 'suitability_gas_cc.sdat',
                             'gas_cc_w_ccs': 'suitability_gas_cc_ccs.sdat',
                             'gas_ct_wo_ccs': 'suitability_gas_cc.sdat',
                             'geothermal': None,
                             'hydro': None,
                             'nuclear_gen_ii': 'suitability_nuclear.sdat',
                             'nuclear_gen_iii': 'suitability_nuclear.sdat',
                             'oil_ct_wo_ccs': 'suitability_oil_baseload.sdat',
                             'solar_csp': 'suitability_solar.sdat',
                             'solar_pv_non_dist': 'suitability_solar.sdat',
                             'wind_onshore': 'suitability_wind.sdat'}

    def __init__(self, settings_dict, lmp_zone_dict, technology_dict, technology_order, infrastructure_dict,
                 initialize_site_data):

        # dictionary containing project level settings
        self.settings_dict = settings_dict

        # dictionary containing lmp zones information
        self.lmp_zone_dict = lmp_zone_dict

        # dictionary containing technology specific information
        self.technology_dict = technology_dict

        # order of technologies to process
        self.technology_order = technology_order

        # infrastructure dictionary
        self.infrastructure_dict = infrastructure_dict

        # initialize model with existing site data
        self.initialize_site_data = initialize_site_data

        # tech_id to tech_name dictionary
        self.tech_name_dict = ({k: self.technology_dict[k].get('tech_name') for k in self.technology_dict.keys()})

        # load coordinate data
        self.cerf_stateid_raster_file = pkg_resources.resource_filename('cerf', 'data/cerf_conus_states_albers_1km.tif')
        self.xcoords, self.ycoords = util.raster_to_coord_arrays(self.cerf_stateid_raster_file)

        # generate grid indices in a flat array
        self.indices_flat = np.array(np.arange(self.xcoords.flatten().shape[0]))
        self.indices_2d = self.indices_flat.reshape(self.xcoords.shape)

        # initialization data for siting
        self.init_arr, self.init_df = self.get_sited_data()

        # raster file containing the lmp zones per grid cell
        self.zones_arr = self.load_lmp_zone_raster()

        # get LMP array per tech [tech_order, x, y]
        logging.info('Processing locational marginal pricing (LMP)')
        self.lmp_arr = self.calculate_lmp()

        # get interconnection cost per tech [tech_order, x, y]
        logging.info('Calculating interconnection costs (IC)')
        self.ic_arr = self.calculate_ic()

        # get NOV array per tech [tech_order, x, y]
        logging.info('Calculating net operational cost (NOV)')
        self.nov_arr = self.calculate_nov()

        # get NLC array per tech [tech_order, x, y]
        logging.info('Calculating net locational cost (NLC)')
        self.nlc_arr = self.calculate_nlc()

        # combine all suitability rasters into an array
        logging.info('Building suitability array')
        self.suitability_arr = self.build_suitability_array()

    def load_lmp_zone_raster(self):
        """Load the lmp zoness raster for the CONUS into a 2D array."""

        # raster file containing the lmp zones per grid cell
        zones_raster_file = self.lmp_zone_dict.get('lmp_zone_raster_file', None)

        # use default if none passed
        if zones_raster_file is None:
            zones_raster_file = pkg_resources.resource_filename('cerf', 'data/lmp_zones_1km.img')

        logging.info(f"Using 'zones_raster_file':  {zones_raster_file}")

        # read in lmp zoness raster as a 2D numpy array
        with rasterio.open(zones_raster_file) as src:
            return src.read(1)

    def calculate_lmp(self):
        """Calculate Locational Marginal Pricing."""

        # create technology specific locational marginal price based on capacity factor
        pricing = LocationalMarginalPricing(self.lmp_zone_dict,
                                            self.technology_dict,
                                            self.technology_order,
                                            self.zones_arr)
        lmp_arr = pricing.get_lmp()

        # get lmp array per tech [tech_order, x, y]
        return lmp_arr

    def calculate_ic(self):
        """Calculate interconnection costs."""

        # unpack configuration and assign defaults
        substation_file = self.infrastructure_dict.get('substation_file', None)
        transmission_costs_file = self.infrastructure_dict.get('transmission_costs_file', None)
        pipeline_costs_file = self.infrastructure_dict.get('pipeline_costs_file', None)
        pipeline_file = self.infrastructure_dict.get('pipeline_file', None)
        output_rasterized_file = self.infrastructure_dict.get('output_rasterized_file', False)
        output_alloc_file = self.infrastructure_dict.get('output_alloc_file', False)
        output_cost_file = self.infrastructure_dict.get('output_cost_file', False)
        output_dist_file = self.infrastructure_dict.get('output_dist_file', False)
        interconnection_cost_file = self.infrastructure_dict.get('interconnection_cost_file', None)

        # instantiate class
        ic = Interconnection(template_array=self.lmp_arr,
                             technology_dict=self.technology_dict,
                             technology_order=self.technology_order,
                             substation_file=substation_file,
                             transmission_costs_file=transmission_costs_file,
                             pipeline_costs_file=pipeline_costs_file,
                             pipeline_file=pipeline_file,
                             output_rasterized_file=output_rasterized_file,
                             output_dist_file=output_dist_file,
                             output_alloc_file=output_alloc_file,
                             output_cost_file=output_cost_file,
                             interconnection_cost_file=interconnection_cost_file,
                             output_dir=self.settings_dict.get('output_directory', None))

        ic_arr = ic.generate_interconnection_costs_array()

        return ic_arr

    def calculate_nov(self):
        """Calculate Net Operational Value."""

        nov_arr = np.zeros_like(self.lmp_arr)

        for index, i in enumerate(self.technology_order):
            econ = NetOperationalValue(discount_rate=self.technology_dict[i]['discount_rate'],
                                       lifetime=self.technology_dict[i]['lifetime'],
                                       unit_size=self.technology_dict[i]['unit_size'],
                                       capacity_factor=self.technology_dict[i]['capacity_factor'],
                                       variable_cost_esc_rate=self.technology_dict[i]['variable_cost_esc_rate'],
                                       fuel_esc_rate=self.technology_dict[i]['fuel_esc_rate'],
                                       carbon_esc_rate=self.technology_dict[i]['carbon_esc_rate'],
                                       variable_om=self.technology_dict[i]['variable_om'],
                                       heat_rate=self.technology_dict[i]['heat_rate'],
                                       fuel_price=self.technology_dict[i]['fuel_price'],
                                       carbon_tax=self.technology_dict[i]['carbon_tax'],
                                       carbon_capture_rate=self.technology_dict[i]['carbon_capture_rate'],
                                       fuel_co2_content=self.technology_dict[i]['fuel_co2_content'],
                                       lmp_arr=self.lmp_arr[index, :, :],
                                       target_year=self.settings_dict.get('run_year'))

            nov_tech_arr = econ.calc_nov()
            nov_arr[index, :, :] = nov_tech_arr

        return nov_arr

    def calculate_nlc(self):
        """Calculate Net Locational Costs."""

        # the most negative number will be the least expensive
        return self.ic_arr - self.nov_arr

    def get_sited_data(self):
        """If initial condition data is provided generate an array to use unsuitable where sites and their buffers
        exists.  Also return a data frame of active sites (not reaching retirement age) to include in the current years
        output.

        """

        # if initial condition data is provided, apply to all technologies
        if self.initialize_site_data is not None:

            # load siting data into a 2D array for the full grid space
            logging.info("Initializing previous siting data")
            init_arr, init_df = util.ingest_sited_data(run_year=self.settings_dict['run_year'],
                                                       x_array=self.xcoords,
                                                       siting_data=self.initialize_site_data)

            return init_arr, init_df

        else:
            return None, None

    def build_suitability_array(self):
        """Build suitability array for all technologies."""

        # set up holder for suitability array
        suitability_array = np.ones_like(self.nlc_arr)

        # load tech specific rasters
        for index, i in enumerate(self.technology_order):

            # path to the input raster
            tech_suitability_raster_file = self.technology_dict[i].get('suitability_raster_file', None)

            if tech_suitability_raster_file is None:
                default_raster = self.DEFAULT_SUITABILITY[self.tech_name_dict[i]]
                tech_suitability_raster_file = pkg_resources.resource_filename('cerf', f'data/{default_raster}')

            logging.info(f"Using suitability file for '{self.technology_dict[i]['tech_name']}':  {tech_suitability_raster_file}")

            # load raster to array
            with rasterio.open(tech_suitability_raster_file) as src:

                # read to 2D array
                tech_arr = src.read(1)

                if self.initialize_site_data is not None:
                    tech_arr = np.maximum(tech_arr, self.init_arr)

                # add to suitability array avoid overwriting the default dimension
                suitability_array[index, :, :] = tech_arr

        return suitability_array




import logging

import numpy as np
import pandas as pd
import rasterio
import xarray as xr
import geopandas as gpd
from shapely.geometry import Point

from cerf.package_data import cerf_crs


def results_to_geodataframe(result_df):
    """Convert the results from 'cerf.run()' to a GeoDataFrame.

    :param result_df:                       Result data frame from running 'cerf.run()'
    :type result_df:                        DataFrame

    :return:                                GeoPandas GeoDataFrame of results

    """

    target_crs = cerf_crs()

    # create geometry column from coordinate fields
    geometry = [Point(xy) for xy in zip(result_df['xcoord'], result_df['ycoord'])]

    return gpd.GeoDataFrame(result_df, crs=target_crs, geometry=geometry)


def kilometers_to_miles(input_km_value):
    """Convert kilometers to miles.

    :param input_km_value:          Kilometer value to convert to miles
    :type input_km_value:           float, int

    :return:                        Miles

    """

    return input_km_value * 0.621371


def suppress_callback(value):
    """Do not log callback output for whitebox functions.

    :param value:                   Value of callback

    """

    pass


def empty_sited_dict():
    """Initialize a sited dictionary."""

    return {'state_name': [],
            'tech_id': [],
            'tech_name': [],
            'xcoord': [],
            'ycoord': [],
            'index': [],
            'buffer_in_km': [],
            'sited_year': [],
            'retirement_year': [],
            'lmp_zone': [],
            'locational_marginal_pricing': [],
            'net_operational_value': [],
            'interconnection_cost': [],
            'net_locational_cost': []}


def sited_dtypes():
    """Return data type dictionary for the sited data frame."""

    return {'state_name': str,
            'tech_id': np.int64,
            'xcoord': np.float64,
            'ycoord': np.float64,
            'lmp_zone': np.int64,
            'locational_marginal_pricing': np.float64,
            'net_operational_value': np.float64,
            'interconnection_cost': np.float64,
            'net_locational_cost': np.float64,
            'index': np.int64,
            'retirement_year': np.int64,
            'sited_year': np.int64,
            'buffer_in_km': np.int64}


def buffer_flat_array(target_index, arr, nrows, ncols, ncells, set_value):
    """Assign a value to the neighboring elements of a 1D array as if they
    were in 2D space. The number of neighbors are based on the `ncells` argument
    which is used to define the window around the target cell to be altered as
    if they were in 2D space.

    :param target_index:                Index of the target element in the 1D array
    :type target_index:                 int

    :param arr:                         A 1D array that has been flattened from a corresponding 2D array
    :type arr:                          ndarray

    :param nrows:                       The number of rows in the parent 2D array
    :type nrows:                        int

    :param ncols:                       The number of columns in the parent 2D array
    :type ncols:                        int

    :param ncells:                      The number of cells for the buffer extending as a radius
    :type ncells:                       int

    :param set_value:                   The value to set for the selected buffer
    :type set_value:                    int; float

    :return:                            [0] Modified 1D array
                                        [1] list of buffered indices

    """
    # list to hold buffer indices for the target grid cell
    buffer_indices = []

    # calculate the number of elements in the 2D grid space
    ngrids = nrows * ncols

    # ensure that the target index is in the grid space
    if 0 <= target_index < ngrids:

        # target cell index bounds for the row
        min_idx = target_index - ncells
        max_idx = target_index + ncells + 1

        # create target row limits
        end_row_idx = target_index + ncols - np.mod(target_index, ncols) - 1
        start_row_idx = end_row_idx - (ncols - 1)

        # do not let the index bleed past the row
        if max_idx > end_row_idx:
            max_idx = end_row_idx + 1

        # do not let the index go negative
        if min_idx < start_row_idx:
            min_idx = start_row_idx

        # initialize above
        min_above = min_idx - ncols
        max_above = max_idx - ncols

        # initialize below
        min_below = min_idx + ncols
        max_below = max_idx + ncols

        # target row assignment
        arr[min_idx: max_idx] = set_value

        # add indices to buffer list
        buffer_indices.extend(list(range(min_idx, max_idx)))

        for _ in range(ncells):

            # above
            if min_above >= 0 and max_above >= 0 and start_row_idx < ngrids:
                arr[min_above: max_above] = set_value

                # add indices to buffer list
                buffer_indices.extend(list(range(min_above, max_above)))

            # advance above to the next row
            min_above -= ncols
            max_above -= ncols

            # below
            if min_below <= ngrids and max_below <= ngrids:
                arr[min_below: max_below] = set_value

                # add indices to buffer list
                buffer_indices.extend(list(range(min_below, max_below)))

            # advance below to the next row
            min_below += ncols
            max_below += ncols

    else:
        raise IndexError(f"Index: '{target_index}' is not in the range of the grid space from 0 to {ngrids - 1}.")

    return arr, buffer_indices


def array_to_raster(arr, template_raster_file, output_raster_file):
    """Write a raster file from a 2D array."""

    with rasterio.open(template_raster_file) as src:

        metadata = src.meta.copy()

        # update metadata
        metadata.update(dtype=np.float64,
                        nodata=np.nan)

        with rasterio.open(output_raster_file, 'w', **metadata) as dest:
            dest.write(arr, 1)


def raster_to_coord_arrays(template_raster):
    """Use the template raster to create two 2D arrays containing the X and Y coordinates of every grid cell.

    :param template_raster:                 Full path with file name and extension to the input raster.
    :type template_raster:                  str

    :return:                                [0] 2D array of X coordinates
                                            [1] 2D array of Y coordinates

    """

    # Read the data
    da = xr.open_rasterio(template_raster)

    # Compute the lon/lat coordinates with rasterio.warp.transform
    x, y = np.meshgrid(da['x'], da['y'])

    return x, y


def ingest_sited_data(run_year, x_array, siting_data):
    """Import sited data containing the locations and additional data to establish an initial suitability condition
    representing power plants and their siting buffer.

    Required fields are the following and they can appear anywhere in the CSV or data frame:

    xcoord:  the X coordinate of the site in meters in USA_Contiguous_Albers_Equal_Area_Conic (EPSG:  102003)
    ycoord:  the Y coordinate of the site in meters in USA_Contiguous_Albers_Equal_Area_Conic (EPSG:  102003)
    retirement_year:  the year (int four digit, e.g., 2050) that the power plant is to be decommissioned
    buffer_in_km:  the buffer around the site to apply in kilometers

    :param run_year:                        Four-digit year of the current run (e.g., 2050)
    :type run_year:                         int

    :param x_array:                         2D array of X coordinates for the entire grid space
    :type x_array:                          ndarray

    :param y_array:                         2D array of Y coordinates for the entire grid space
    :type y_array:                          ndarray

    :param siting_data:                     Full path with file name and extension for the input siting file or a
                                            Pandas DataFrame
    :type siting_data:                      str, DataFrame

    :return:                                [0] 2D array of 0 (suitable) and 1 (unsuitable) values where 1 are the sites
                                            and their buffers of active power plants

                                            [1] Pandas DataFrame of active sites (not retired)

    """

    # assign input data to a data frame
    if isinstance(siting_data, pd.DataFrame):
        df = siting_data
    elif isinstance(siting_data, str):
        df = pd.read_csv(siting_data, dtype=sited_dtypes())
    else:
        msg = "The user must pass either a CSV file path to 'sited_csv' or a Pandas DataFrame to 'sited_df'"
        logging.error(msg)
        raise TypeError()

    # only keep sites that are not retired
    df_active = df.loc[df['retirement_year'] > run_year].copy()

    # initialize an array to hold the 0, 1 sited and buffer data
    sited_arr = np.zeros_like(x_array).flatten()

    for ix in df_active['index'].tolist():

        # get the buffer size for the site
        site_buffer_km = df_active.loc[df_active['index'] == ix]['buffer_in_km'].values[0]

        # apply the buffer to the site and set to the entire array
        sited_arr = buffer_flat_array(ix, sited_arr, x_array.shape[0], x_array.shape[1], site_buffer_km, 1)[0]

    return sited_arr.reshape(x_array.shape).astype(np.int8), df_active



import logging
import os

import pkg_resources
import yaml

from cerf.logger import Logger


class ReadConfig(Logger):
    """Read the configuration YAML file to a dictionary. Users can optionally pass in a configuration dictionary
    instead.

        :param config_file:                 Full path with file name and extension to the input config.yml file
        :type config_file:                  str

        :param config_dict:                 Configuration dictionary.
        :type config_dict:                  dict

    """

    # type hints
    config_file: str

    def __init__(self, config_file=None, config_dict={}):

        # inherit logger class attributes
        super(ReadConfig, self).__init__()

        if config_file is None and config_dict is not None:
            self.config = config_dict

        else:

            # yaml config file
            self.config_file = config_file

            # read into dict
            self.config = self.get_yaml()

        # get project level settings
        self.settings_dict = self.config.get('settings')
        self.settings_dict.update(config_dict.get('settings', {}))

        # generate the technology order that will be use for indexing arrays throughout modeling
        self.technology_dict = self.config.get('technology')
        self.technology_dict.update(config_dict.get('technology', {}))

        self.technology_order = list(self.technology_dict.keys())

        # get the expansion plan
        self.expansion_dict = self.config.get('expansion_plan')
        self.expansion_dict.update(config_dict.get('expansion_plan', {}))

        # get the lmp zones settings
        self.lmp_zone_dict = self.config.get('lmp_zones')
        self.lmp_zone_dict.update(config_dict.get('lmp_zones', {}))

        # get the infrastructure settings
        self.infrastructure_dict = self.config.get('infrastructure', {})
        self.infrastructure_dict.update(config_dict.get('infrastructure', {}))

        # get the states dictionary
        self.states_dict = self.get_states_dict()

    @staticmethod
    def read_yaml(yaml_file):
        """Read a YAML file."""

        with open(yaml_file, 'r') as yml:
            return yaml.load(yml, Loader=yaml.FullLoader)

    def get_yaml(self):
        """Read the YAML config file.

        :return:                            YAML config object

        """

        # if config file not passed
        if self.config_file is None:
            msg = "Config file must be passed as an argument using:  config_file='<path to config.yml'>"
            logging.error(msg)
            raise AttributeError(msg)

        # check for path exists
        if os.path.isfile(self.config_file):

            return self.read_yaml(self.config_file)

        else:
            msg = f"Config file not found for path:  {self.config_file}"
            logging.error(msg)
            raise FileNotFoundError(msg)

    def get_states_dict(self):
        """Get a dictionary of state name to state ID from the YAML file in package data."""

        # in package data {state_name: state_id}
        states_lookup_file = pkg_resources.resource_filename('cerf', 'data/state-name_to_state-id.yml')

        return self.read_yaml(states_lookup_file)



import re
from setuptools import setup, find_packages


def readme():
    """Return the contents of the project README file."""
    with open('README.md') as f:
        return f.read()


def get_requirements():
    """Return a list of package requirements from the requirements.txt file."""
    with open('requirements.txt') as f:
        return f.read().split()


version = re.search(r"__version__ = ['\"]([^'\"]*)['\"]", open('cerf/__init__.py').read(), re.M).group(1)

setup(
    name='cerf',
    version=version,
    packages=find_packages(),
    url='https://github.com/IMMM-SFA/cerf',
    download_url=f'https://github.com/IMMM-SFA/cerf/archive/refs/tags/v{version}.tar.gz',
    license='BSD2-Simplified',
    author='Chris R. Vernon; Nino Zuljevic',
    author_email='chris.vernon@pnnl.gov',
    description='An open-source geospatial Python package for assessing and analyzing future electricity technology capacity expansion feasibility',
    long_description=readme(),
    long_description_content_type="text/markdown",
    python_requires='>=3.7.*, <4',
    include_package_data=True,
    install_requires=[
        'numpy~=1.19.4',
        'pandas~=1.1.4',
        'rasterio~=1.2.3',
        'xarray~=0.16.1',
        'PyYAML~=5.4.1',
        'requests~=2.25.1',
        'joblib~=1.0.1',
        'matplotlib~=3.3.3',
        'seaborn~=0.11.1',
        'whitebox~=1.5.1',
        'fiona~=1.8.19',
        'pyproj~=3.0.1',
        'rtree~=0.9.7',
        'shapely~=1.7.1',
        'geopandas~=0.9.0'
    ],
    extras_require={
        'dev': [
            'build~=0.5.1',
            'nbsphinx~=0.8.6',
            'setuptools~=57.0.0',
            'sphinx~=4.0.2',
            'sphinx-panels~=0.6.0',
            'sphinx-rtd-theme~=0.5.2',
            'twine~=3.4.1'
        ]
    }
)



import os
import logging
import pkg_resources
import tempfile

import rasterio
import whitebox
import yaml

import numpy as np
import geopandas as gpd

from rasterio import features

from cerf.utils import suppress_callback
from cerf.package_data import cerf_crs, costs_per_kv_substation, get_state_abbrev_to_name

# instantiate whitebox toolset
wbt = whitebox.WhiteboxTools()


class Interconnection:
    """Calculate interconnection costs per grid cell in $ / yr using:

    Interconnection Cost ($ / yr) = Distance to nearest suitable transmission line (km) *
                                        Electric grid interconnection captial cost ($ / km) *
                                        Annuity factor
                                        + (if gas-fired technology) Distance to nearest suitable gas pipeline (km) *
                                        Gas interconnection captial cost ($ / km) *
                                        Annuity factor

            where, Annuity factor is (d(1 + d)**n) / ((1 + d)**n - 1)
            where, d = real annual discount rate (%), n = asset lifetime (years)


    :param technology_dict:                 Dictionary containing technology specific information from the config file
    :type technology_dict:                  dict

    :param technology_order:                Order of technologies to process
    :type technology_order:                 list

    :param substation_file:                 Full path with file name and extension to the input substations shapefile.
                                            If None, CERF will use the default data stored in the package.

    :type substation_file:                  str

    :param transmission_costs_dict:         A dictionary containing the cost of connection per km to a substation
                                            having a certain minimum voltage range.  Default is to load from the
                                            CERF data file 'costs_per_kv_substation.yml' by specifying 'None'

    :type transmission_costs_dict:          dict

    :param transmission_costs_file:         A YAML file containing the cost of connection per km to a substation
                                            having a certain minimum voltage range.  Default is to load from the
                                            CERF data file 'costs_per_kv_substation.yml' by specifying 'None'

    :type transmission_costs_file:          str

    :param pipeline_costs_dict:             A dictionary containing the cost of connection per km to a gas pipeline.
                                            Default is to load from the CERF data file 'costs_gas_pipeline.yml' by
                                            specifying 'None'

    :type pipeline_costs_dict:              dict

    :param pipeline_costs_file:             A YAML file containing the cost of connection per km to a gas pipeline.
                                            Default is to load from the CERF data file 'costs_gas_pipeline.yml' by
                                            specifying 'None'

    :type pipeline_costs_file:              str


    :param pipeline_file:                   Full path with file name and extension to the input pipelines shapefile.
                                            If None, CERF will use the default data stored in the package.

    :type pipeline_file:                    str

    :param output_rasterized_file:          Write distance raster; if True, set 'output_dir' value
    :type output_rasterized_file:           bool

    :param output_dist_file:                Write distance raster; if True, set 'output_dir' value
    :type output_dist_file:                 bool

    :param output_alloc_file:               Write allocation file; if True, set 'output_dir' value
    :type output_alloc_file:                bool

    :param output_cost_file:                Write cost file; if True, set 'output_dir' value
    :type output_cost_file:                 bool
    
    :param interconnection_cost_file:       Full path with file name and extension to a preprocessed interconnection
                                            cost NPY file that has been previously written. If None, IC will be 
                                            calculated.
    :type interconnection_cost_file:        str 

    :param output_dir:                      Full path to a directory to write outputs to if desired
    :type output_dir:                       str

    """

    def __init__(self, template_array, technology_dict, technology_order, substation_file=None,
                 transmission_costs_dict=None, transmission_costs_file=None, pipeline_costs_dict=None, 
                 pipeline_costs_file=None,pipeline_file=None, output_rasterized_file=False, output_dist_file=False, 
                 output_alloc_file=False, output_cost_file=False, interconnection_cost_file=None, output_dir=None):

        self.template_array = template_array
        self.technology_dict = technology_dict
        self.technology_order = technology_order
        self.substation_file = substation_file
        self.transmission_costs_dict = transmission_costs_dict
        self.transmission_costs_file = transmission_costs_file
        self.pipeline_costs_dict = pipeline_costs_dict
        self.pipeline_costs_file = pipeline_costs_file
        self.pipeline_file = pipeline_file
        self.output_rasterized_file = output_rasterized_file
        self.output_dist_file = output_dist_file
        self.output_alloc_file = output_alloc_file
        self.output_cost_file = output_cost_file
        self.interconnection_cost_file = interconnection_cost_file
        self.output_dir = output_dir

        # calculate electricity transmission infrastructure costs
        self.substation_costs = self.transmission_to_cost_raster(setting='substations')

        # if there are any gas technlogies present, calculate gas pipeline infrastructure costs
        self.pipeline_costs = self.transmission_to_cost_raster(setting='pipelines')

    @staticmethod
    def calc_annuity_factor(discount_rate, lifetime):
        """Calculate annuity factor."""

        fx = pow(1.0 + discount_rate, lifetime)

        return discount_rate * fx / (fx - 1.0)

    def get_pipeline_costs(self):
        """Get the costs of gas pipeline interconnection per kilometer."""

        if self.pipeline_costs_dict is not None:
            logging.info(f"Using gas pipeline costs from user defined dictionary:  {self.pipeline_costs_dict}")
            return self.pipeline_costs_dict.get('gas_pipeline_cost')

        if self.pipeline_costs_file is not None:
            f = self.pipeline_costs_file
            logging.info(f"Using gas pipeline costs from file:  {f}")

        else:
            f = pkg_resources.resource_filename('cerf', 'data/costs_gas_pipeline.yml')
            logging.info(f"Using gas pipeline costs from default file:  {f}")

        with open(f, 'r') as yml:
            yaml_dict = yaml.load(yml, Loader=yaml.FullLoader)

        return yaml_dict.get('gas_pipeline_cost')

    def process_hifld_substations(self):
        """Select substations from HIFLD data that are within the CONUS and either in service or under construction and
        having a minimum voltage rating >= 0.  A field used to rasterize ('_rval_') is also added containing the cost of
        connection in $/km for each substation.

        This data is assumed to have the following fields:  ['TYPE', 'STATE', 'STATUS'].

        :returns:                               A geodataframe containing the target substations

        """

        # load cost dictionary from package data if none passed
        if (self.transmission_costs_dict is None) and (self.transmission_costs_file is None):
            default_kv_file = pkg_resources.resource_filename('cerf', 'data/costs_per_kv_substation.yml')

            logging.info(f"Using default substations costs from file: {default_kv_file}")

            self.transmission_costs_dict = costs_per_kv_substation()

        elif self.transmission_costs_file is not None:

            logging.info(f"Using substations costs from file: {self.transmission_costs_file}")

            with open(self.transmission_costs_file, 'r') as yml:
                self.transmission_costs_dict = yaml.load(yml, Loader=yaml.FullLoader)

        if self.substation_file is None:

            sub_file = pkg_resources.resource_filename('cerf', 'data/hifld_substations_conus_albers.zip')

            logging.info(f"Using default substation file: {sub_file}")

            return gpd.read_file(sub_file)

        else:

            logging.info(f"Using substation file: {self.substation_file}")

            # get state abbreviations file from cerf package data
            states = get_state_abbrev_to_name()

            # load cerf's default coordinate reference system object
            target_crs = cerf_crs()

            # load and reproject
            gdf = gpd.read_file(self.substation_file).to_crs(target_crs)

            # keep only substations in the CONUS that are either in service or under construction
            gdf = gdf.loc[(gdf['TYPE'] == 'SUBSTATION') &
                          (gdf['STATE'].isin(states.keys())) &
                          (gdf['STATUS'].isin(('IN SERVICE', 'UNDER CONST')))].copy()

            # assign a field to rasterize by containing the cost of transmission per km
            gdf['_rval_'] = 0

            for i in self.transmission_costs_dict.keys():
                gdf['_rval_'] = np.where((gdf['MIN_VOLT'] >= self.transmission_costs_dict[i]['min_voltage']) &
                                         (gdf['MIN_VOLT'] <= self.transmission_costs_dict[i]['max_voltage']),
                                         self.transmission_costs_dict[i]['dollar_per_km'],
                                         gdf['_rval_'])

            return gdf

    def process_eia_natural_gas_pipelines(self):
        """Select natural gas pipelines from EIA data that have a status of operating and a length greater than 0.

        :returns:                               A geodataframe containing the target pipelines

        """

        if self.pipeline_file is None:

            f = pkg_resources.resource_filename('cerf', 'data/eia_natural_gas_pipelines_conus_albers.zip')

            logging.info(f"Using default gas pipeline file:  {f}")

            # read in default shapefile for pipelines
            gdf = gpd.read_file(f)

            # set field for rasterize
            gdf['_rval_'] = self.get_pipeline_costs()

            return gdf

        else:

            logging.info(f"Using gas pipeline file:  {self.pipeline_file}")

            # load cerf's default coordinate reference system object
            target_crs = cerf_crs()

            # read in data and reproject
            gdf = gpd.read_file(self.pipeline_file).to_crs(target_crs)

            # only keep features with a length > 0
            gdf = gdf.loc[gdf.geometry.length > 0].copy()

            # only keep operational pipelines
            gdf = gdf.loc[gdf['Status'] == 'Operating'].copy()

            # set field for rasterize
            gdf['_rval_'] = self.get_pipeline_costs()

            return gdf

    def transmission_to_cost_raster(self, setting):
        """Create a cost per grid cell in $/km from the input GeoDataFrame of transmission infrastructure having a cost
        designation field as '_rval_'

        :param setting:                         Either 'substations' or 'pipelines'
        :type setting:                          str

        :return:                                Array of transmission interconnection cost per grid cell

        """
        # conversion factor for meters to km
        m_to_km_factor = 0.001

        if setting == 'substations':
            gdf = self.process_hifld_substations()

        elif setting == 'pipelines':
            gdf = self.process_eia_natural_gas_pipelines()

        else:
            raise ValueError(f"Incorrect setting '{setting}' for transmission data.  Must be 'substations' or 'pipelines'")

        # get the template raster from CERF data
        template_raster = pkg_resources.resource_filename('cerf', 'data/cerf_conus_states_albers_1km.tif')

        with rasterio.open(template_raster) as src:
            # create 0 where land array
            arr = (src.read(1) * 0).astype(rasterio.float64)

            # update metadata datatype to float64
            metadata = src.meta.copy()
            metadata.update({'dtype': rasterio.float64})

            # reproject transmission data
            infrastructure_gdf = gdf.to_crs(src.crs)

            # get shapes
            shapes = ((geom, value) for geom, value in zip(infrastructure_gdf.geometry, infrastructure_gdf['_rval_']))

        with tempfile.TemporaryDirectory() as tempdir:

            # if write desired
            if any((self.output_rasterized_file, self.output_dist_file, self.output_alloc_file, self.output_cost_file)):

                if self.output_dir is None:
                    msg = "If writing rasters to file must specify 'output_dir'"
                    logging.error(msg)
                    raise NotADirectoryError(msg)

                else:
                    out_rast = os.path.join(self.output_dir, f'cerf_transmission_raster_{setting}.tif')
                    out_dist = os.path.join(self.output_dir, f'cerf_transmission_distance_{setting}.tif')
                    out_alloc = os.path.join(self.output_dir, f'cerf_transmission_allocation_{setting}.tif')
                    out_costs = os.path.join(self.output_dir, f'cerf_transmission_costs_{setting}.tif')
            else:
                out_rast = os.path.join(tempdir, f'cerf_transmission_raster_{setting}.tif')
                out_dist = os.path.join(tempdir, f'cerf_transmission_distance_{setting}.tif')
                out_alloc = os.path.join(tempdir, f'cerf_transmission_allocation_{setting}.tif')
                out_costs = os.path.join(tempdir, f'cerf_transmission_costs_{setting}.tif')

            # rasterize transmission vector data and write to memory
            with rasterio.open(out_rast, 'w', **metadata) as dataset:
                # burn features into raster
                burned = features.rasterize(shapes=shapes, fill=0, out=arr, transform=dataset.transform)

                # write the outputs to file
                dataset.write_band(1, burned)

            # calculate Euclidean distance and write raster; result just stores the return value 0
            dist_result = wbt.euclidean_distance(out_rast, out_dist, callback=suppress_callback)

            # calculate Euclidean allocation and write raster
            alloc_result = wbt.euclidean_allocation(out_rast, out_alloc, callback=suppress_callback)

            with rasterio.open(out_dist) as dist:
                dist_arr = dist.read(1)

            with rasterio.open(out_alloc) as alloc:
                alloc_arr = alloc.read(1)

            with rasterio.open(out_costs, 'w', **metadata) as out:

                # distance in km * the cost of the nearest substation; outputs $/km
                cost_arr = (dist_arr * m_to_km_factor) * alloc_arr

                out.write(cost_arr, 1)

        return cost_arr

    def generate_interconnection_costs_array(self):
        """Calculate the costs of interconnection for each technology."""
        
        # if a preprocessed file has been provided, load and return it
        if self.interconnection_cost_file is not None:

            logging.info(f"Using prebuilt interconnection costs file:  {self.interconnection_cost_file}")
            return np.load(self.interconnection_cost_file)

        # set up array to hold interconnection costs
        ic_arr = np.zeros_like(self.template_array)

        for index, i in enumerate(self.technology_order):

            # get technology specific information
            require_pipelines = self.technology_dict[i].get('require_pipelines', False)
            discount_rate = self.technology_dict[i].get('discount_rate')
            lifetime = self.technology_dict[i].get('lifetime')

            # calulate annuity factor for technology
            annuity_factor = self.calc_annuity_factor(discount_rate=discount_rate, lifetime=lifetime)

            # get transmission cost array
            substation_cost_array = self.substation_costs * annuity_factor

            if require_pipelines:

                # get pipeline cost array
                pipeline_cost_array = self.pipeline_costs * annuity_factor

                # calculate technology specific interconnection cost
                total_interconection_cost_array = substation_cost_array + pipeline_cost_array

            else:
                total_interconection_cost_array = substation_cost_array

            # calculate interconnection costs per grid cell
            ic_arr[index, :, :] = total_interconection_cost_array

        return ic_arr




import matplotlib.pyplot as plt

from cerf.package_data import cerf_states_shapefile, cerf_boundary_shapefile
from cerf.utils import results_to_geodataframe


def plot_siting(result_df, column='tech_name', markersize=5, cmap='Paired', save_figure=False, output_file=None):
    """Plot the results of a cerf run on a map where each technology has its own color.

    :param result_df:                       Result data frame from running 'cerf.run()'
    :type result_df:                        DataFrame

    :param column:                          Column to plot
    :type column:                           str

    :param markersize:                      Size of power plant marker
    :type markersize:                       int

    :param cmap:                            Custom matplotlib colormap object or name

    :param save_figure:                     If True, figure is saved to file and 'output_file' must be set
    :type save_figure:                      bool

    :param output_file:                     If 'save_figure' is True, specify full path with file name and extension
                                            for the file to be saved to

    """

    # result df to geodataframe
    gdf = results_to_geodataframe(result_df)

    fig, ax = plt.subplots(figsize=(20, 10))

    # read in boundary data
    states_gdf = cerf_states_shapefile()
    boundary_gdf = cerf_boundary_shapefile()

    # add background
    boundary_gdf.plot(ax=ax, color="#f3f2f2", lw=0.8)

    # add boundaries
    states_gdf.boundary.plot(ax=ax, edgecolor='gray', lw=0.2)
    boundary_gdf.boundary.plot(ax=ax, edgecolor='gray', lw=0.8)

    # add data
    gdf.plot(ax=ax, markersize=markersize, column=column, cmap=cmap, legend=True)

    ax.set_axis_off()

    if save_figure is True and output_file is None:
        raise ValueError(f"'output_file' must be set if 'save_figure' is True.")

    if save_figure:
        plt.savefig(output_file)



"""Process a state for the target year.

@author Chris R. vernon
@email chris.vernon@pnnl.gov

License:  BSD 2-Clause, see LICENSE and DISCLAIMER files

"""

import logging
import os
import time

import numpy as np
import pandas as pd
import pkg_resources
import rasterio

from cerf.compete import Competition


class ProcessState:

    def __init__(self,
                 settings_dict,
                 technology_dict,
                 technology_order,
                 expansion_dict,
                 states_dict,
                 suitability_arr,
                 lmp_arr,
                 nov_arr,
                 ic_arr,
                 nlc_arr,
                 zones_arr,
                 xcoords,
                 ycoords,
                 indices_2d,
                 target_state_name,
                 randomize=True,
                 seed_value=0,
                 verbose=False,
                 write_output=False):

        # dictionary containing project level settings
        self.settings_dict = settings_dict

        # dictionary containing technology specific information
        self.technology_dict = technology_dict

        # order of technologies to process
        self.technology_order = technology_order

        # dictionary containing the expansion plan
        self.expansion_dict = expansion_dict

        # states dictionary with state name to state ID mapping
        self.states_dict = states_dict

        # target state name
        self.target_state_name = target_state_name

        # the id of the target state as it is represented in the state raster
        self.target_state_id = self.get_state_id()

        # suitability data for the CONUS
        self.suitability_arr = suitability_arr

        # LMP array for the CONUS
        self.lmp_arr = lmp_arr

        # NOV array for the CONUS
        self.nov_arr = nov_arr

        # IC array for the CONUS
        self.ic_arr = ic_arr

        # NLC data for the CONUS
        self.nlc_arr = nlc_arr

        # lmp zoness for the CONUS
        self.zones_arr = zones_arr

        # coordinates for each index
        self.xcoords = xcoords
        self.ycoords = ycoords

        # the choice to randomize when a technology has more than one NLC cheapest value
        self.randomize = randomize

        # a random seed value that is used when the user wants to replicate a run exactly
        self.seed_value = seed_value

        # log verbose siting information
        self.verbose = verbose

        # set write outputs flag
        self.write_outputs = write_output

        logging.debug(f"Extracting suitable grids for {self.target_state_name}")
        self.suitability_array_state, self.ymin, self.ymax, self.xmin, self.xmax = self.extract_state_suitability()

        logging.debug(f"Creating a NLC state level array for {self.target_state_name}")
        self.suitable_nlc_state = self.mask_nlc()

        logging.debug(f"Generating grid indices for {self.target_state_name}")
        # grid indices for the entire grid in a 2D array
        self.indices_2d = indices_2d
        self.indices_flat_state = self.get_grid_indices()

        logging.debug(f"Get grid coordinates for {self.target_state_name}")
        self.xcoords_state, self.ycoords_state = self.get_grid_coordinates()

        logging.debug(f"Extracting additional metrics for {self.target_state_name}")
        self.lmp_flat_dict, self.nov_flat_dict, self.ic_flat_dict = self.extract_state_metrics()
        self.zones_flat_arr = self.extract_lmp_zones()

        logging.debug(f"Competing technologies to site expansion for {self.target_state_name}")
        self.run_data = self.competition()

    def get_state_id(self):
        """Load state name to state id YAML file to a dictionary.

        :return:                        Corresponding state ID for the user passed state name.

        """

        if self.target_state_name in self.states_dict:
            return self.states_dict.get(self.target_state_name.lower())

        else:

            logging.error(f"State name: `{self.target_state_name}` not in registry.")
            logging.error(f"Please select a state name from the following:  {list(self.states_dict.keys())}")

            raise KeyError()

    def extract_state_suitability(self):
        """Extract a single state from the suitability."""

        # load the state raster as array
        state_raster_file = pkg_resources.resource_filename('cerf', 'data/cerf_conus_states_albers_1km.tif')
        with rasterio.open(state_raster_file) as src:
            states_arr = src.read(1)

        # get target state indices in grid space
        state_indices = np.where(states_arr == self.target_state_id)

        # get minimum and maximum bounds
        ymin = np.min(state_indices[0])
        ymax = np.max(state_indices[0]) + 1
        xmin = np.min(state_indices[1])
        xmax = np.max(state_indices[1]) + 1

        # extract state and give binary designation
        state_mask = states_arr[ymin:ymax, xmin:xmax].copy()
        state_mask = np.where(state_mask == self.target_state_id, 0, 1)

        # extract state footprint from suitability data
        suitability_array_state = self.suitability_arr[:, ymin:ymax, xmin:xmax].copy()

        # add in suitability where unsuitable is the highest value of NLC
        suitability_array_state += state_mask

        # at this point, we have all suitable grid cells as 0 and all not as 1
        suitability_array_state = np.where(suitability_array_state == 0, 0, 1)

        # exclude all area for the default dimension
        suitability_array_state = np.insert(suitability_array_state, 0, np.ones_like(suitability_array_state[0, :, :]), axis=0)

        return suitability_array_state, ymin, ymax, xmin, xmax

    def mask_nlc(self):
        """Extract NLC elements for the current state."""

        # extract state footprint from NLC data
        nlc_arr_state = self.nlc_arr[:, self.ymin:self.ymax, self.xmin:self.xmax].copy()

        # insert zero array, mask it as index [0, :, :] so the tech_id 0 will always be min if nothing is left to site
        nlc_arr_state = np.insert(nlc_arr_state, 0, np.zeros_like(nlc_arr_state[0, :, :]), axis=0)

        # make any nan grid cells the most expensive option to exclude
        nlc_arr_state = np.nan_to_num(nlc_arr_state, nan=np.nanmax(nlc_arr_state) + 1)

        # apply the mask to NLC data
        return np.ma.masked_array(nlc_arr_state, mask=self.suitability_array_state)

    def get_grid_indices(self):
        """Generate a 1D array of grid indices the target state to use as a way to map state level outcomes back to the
        full grid space."""

        return self.indices_2d[self.ymin:self.ymax, self.xmin:self.xmax].flatten()

    def get_grid_coordinates(self):
        """Generate 1D arrays of grid coordinates (X, Y) to use for siting based on the bounds of the target state."""

        xcoord_2d_state = self.xcoords[self.ymin:self.ymax, self.xmin:self.xmax].flatten()
        ycoord_2d_state = self.ycoords[self.ymin:self.ymax, self.xmin:self.xmax].flatten()

        return xcoord_2d_state, ycoord_2d_state

    def extract_state_metrics(self):
        """Extract the LMP, NOV, and IC arrays for the target state and return them as dictionaries where
        {tech_id: flat_array, ...}.

        """

        # extract the target state
        lmp_arr_state = self.lmp_arr[:, self.ymin:self.ymax, self.xmin:self.xmax]
        nov_arr_state = self.nov_arr[:, self.ymin:self.ymax, self.xmin:self.xmax]
        ic_arr_state = self.ic_arr[:, self.ymin:self.ymax, self.xmin:self.xmax]

        # create a reference dictionary where {tech_id: flat_state_array, ...}
        lmp_flat_dict = {i: lmp_arr_state[ix, :, :].flatten() for ix, i in enumerate(self.technology_order)}
        nov_flat_dict = {i: nov_arr_state[ix, :, :].flatten() for ix, i in enumerate(self.technology_order)}
        ic_flat_dict = {i: ic_arr_state[ix, :, :].flatten() for ix, i in enumerate(self.technology_order)}

        return lmp_flat_dict, nov_flat_dict, ic_flat_dict

    def extract_lmp_zones(self):
        """Extract the lmp zoness elements for the target state and return as a flat array."""

        return self.zones_arr[self.ymin:self.ymax, self.xmin:self.xmax].flatten()

    def competition(self):
        """Compete technologies."""

        comp = Competition(target_state_name=self.target_state_name,
                           settings_dict=self.settings_dict,
                           technology_dict=self.technology_dict,
                           technology_order=self.technology_order,
                           expansion_dict=self.expansion_dict[self.target_state_name],
                           lmp_dict=self.lmp_flat_dict,
                           nov_dict=self.nov_flat_dict,
                           ic_dict=self.ic_flat_dict,
                           nlc_mask=self.suitable_nlc_state,
                           zones_arr=self.zones_flat_arr,
                           xcoords=self.xcoords_state,
                           ycoords=self.ycoords_state,
                           indices_flat=self.indices_flat_state,
                           randomize=self.randomize,
                           seed_value=self.seed_value,
                           verbose=self.verbose)

        # create data frame of sited data
        df = pd.DataFrame(comp.sited_dict)

        # write outputs if so desired
        if self.write_outputs:

            # create output CSV file of coordinate data
            csv_file_name = f"cerf_sited_{self.settings_dict['run_year']}_{self.target_state_name}.csv"
            csv_out_file = os.path.join(self.settings_dict.get('output_directory'), csv_file_name)

            df.to_csv(csv_out_file, index=False)

        return comp


def process_state(target_state_name, settings_dict, technology_dict, technology_order, expansion_dict, states_dict,
                  suitability_arr, lmp_arr, nov_arr, ic_arr, nlc_arr, zones_arr, xcoords, ycoords, indices_2d,
                  randomize=True, seed_value=0, verbose=False, write_output=True):
    """Convenience wrapper to log time and site an expansion plan for a target state for the target year.

    :param target_state_name:                   Name of the target state as it is represented in the state raster.
                                                Must be all lower case with spacing separated by an underscore.
    :type target_state_name:                    str

    :param settings_dict:                       Project level setting dictionary from cerf.read_config.ReadConfig
    :type settings_dict:                        dict

    :param technology_dict:                     Technology level data dictionary from cerf.read_config.ReadConfig
    :type technology_dict:                      dict

    :param technology_order:                    Technology processing order to index by from cerf.read_config.ReadConfig
    :type technology_order:                     list

    :param expansion_dict:                      Expansion plan data dictionary from cerf.read_config.ReadConfig
    :type expansion_dict:                       dict

    :param states_dict:                         Mapping from state name to state ID from cerf.read_config.ReadConfig
    :type states_dict:                          dict

    :param suitability_arr:                     3D array where {tech_id, x, y} for suitability data
    :type suitability_arr:                      ndarray

    :param nlc_arr:                             3D array where {tech_id, x, y} for NLC data
    :type nlc_arr:                              ndarray

    :param data:                                Object containing all data (NLC, etc.) to run the expansion. This
                                                data is generated from the cerf.stage.Stage class.
    :type data:                                 class

    :param randomize:                           Choice to randomize when a technology has more than one NLC
                                                cheapest value
    :type randomize:                            bool

    :param seed_value:                          A random seed value that is used when the user wants to replicate
                                                a run exactly
    :type seed_value:                           int

    :param verbose:                             Log verbose siting information
    :type verbose:                               bool

    :param write_output:                        Choice to write output to a file
    :type write_output:                         bool

    :return:                                    2D NumPy array of sited technologies in the CONUS grid space where
                                                grid cell values are in the technology number as provided by the
                                                expansion plan

    """

    logging.debug(f'Processing state:  {target_state_name}')

    # check to see if state has any sites in the expansion
    n_sites = sum([expansion_dict[target_state_name][k]['n_sites'] for k in expansion_dict[target_state_name].keys()])

    # if there are no sites in the expansion, return an all NaN 2D array
    if n_sites <= 0:
        logging.warning(f"There were no sites expected for any technology in `{target_state_name}`")
        return None

    else:

        # initial time for processing state
        state_t0 = time.time()

        # process expansion plan and competition for a single state for the target year
        process = ProcessState(settings_dict=settings_dict,
                               technology_dict=technology_dict,
                               technology_order=technology_order,
                               expansion_dict=expansion_dict,
                               states_dict=states_dict,
                               suitability_arr=suitability_arr,
                               lmp_arr=lmp_arr,
                               nov_arr=nov_arr,
                               ic_arr=ic_arr,
                               nlc_arr=nlc_arr,
                               zones_arr=zones_arr,
                               xcoords=xcoords,
                               ycoords=ycoords,
                               indices_2d=indices_2d,
                               target_state_name=target_state_name,
                               randomize=randomize,
                               seed_value=seed_value,
                               verbose=verbose,
                               write_output=write_output)

        logging.info(f'Processed `{target_state_name}` in {round(time.time() - state_t0, 7)} seconds')

        return process



import unittest

import numpy as np
import pkg_resources
import rasterio

from cerf.lmp import LocationalMarginalPricing
from cerf.read_config import ReadConfig
from cerf.package_data import config_file


class TestLmp(unittest.TestCase):

    # supporting data
    SLIM_LMP_ARRAY = np.load(pkg_resources.resource_filename('cerf', 'tests/data/comp_data/lmp_arr.npy'))
    LMP_FILE = pkg_resources.resource_filename('cerf', 'data/illustrative_lmp_8760-per-zone_dollars-per-mwh.zip')
    TEST_CONFIG = pkg_resources.resource_filename('cerf', 'tests/data/test_config_2010.yml')

    @staticmethod
    def get_sample(arr):
        """Get a sample from the LMP dictionary to reduce comparison size.  Sample space covers the Southeast U.S."""

        return arr[:, 1500:2000, 3000:4000].copy()

    @staticmethod
    def load_lmp_zone_raster(lmp_zone_dict):
        """Load the lmp zoness raster for the CONUS into a 2D array."""

        # raster file containing the lmp zones per grid cell
        zones_raster_file = lmp_zone_dict.get('lmp_zone_raster_file')

        if zones_raster_file is None:
            zones_raster_file = pkg_resources.resource_filename('cerf', 'data/lmp_zones_1km.img')

        # read in lmp zoness raster as a 2D numpy array
        with rasterio.open(zones_raster_file) as src:
            return src.read(1)

    def test_lmp_outputs(self):
        """Test to make sure LMP outputs match expected."""

        # set seed for reproducibility
        np.random.seed(0)

        # read in configuration file
        cfg = ReadConfig(self.TEST_CONFIG)

        # read in zones array
        zones_arr = self.load_lmp_zone_raster(cfg.lmp_zone_dict)

        # create technology specific locational marginal price based on capacity factor
        pricing = LocationalMarginalPricing(cfg.lmp_zone_dict,
                                            cfg.technology_dict,
                                            cfg.technology_order,
                                            zones_arr)

        # get lmp array per tech [tech_order, x, y]
        lmp_arr = pricing.get_lmp()

        # trim down LMP array for testing
        slim_lmps = self.get_sample(lmp_arr)

        # test LMP array equality
        np.testing.assert_array_equal(np.around(TestLmp.SLIM_LMP_ARRAY, 4), np.around(slim_lmps, 4))


if __name__ == '__main__':
    unittest.main()



"""Model interface for CERF

@author Chris R. vernon
@email chris.vernon@pnnl.gov

License:  BSD 2-Clause, see LICENSE and DISCLAIMER files

"""


import logging
import time

from cerf.process_state import process_state
from cerf.read_config import ReadConfig
from cerf.stage import Stage


class Model(ReadConfig):
    """Model wrapper for CERF.

    :param config_file:                 Full path with file name and extension to the input config.yml file
    :type config_file:                  str

    :param config_dict:                 Optional instead of config_file. Configuration dictionary.
    :type config_dict:                  dict

    :param initialize_site_data:        None if no initialization is required, otherwise either a CSV file or
                                        Pandas DataFrame of siting data bearing the following required fields:

                                        xcoord:  the X coordinate of the site in meters in
                                        USA_Contiguous_Albers_Equal_Area_Conic (EPSG:  102003)

                                        ycoord:  the Y coordinate of the site in meters in
                                        USA_Contiguous_Albers_Equal_Area_Conic (EPSG:  102003)

                                        retirement_year:  the year (int four digit, e.g., 2050) that the power
                                        plant is to be decommissioned

                                        buffer_in_km:  the buffer around the site to apply in kilometers

    :param log_level:                   Log level.  Options are 'info' and 'debug'.  Default 'info'
    :type log_level:                    str

    """

    def __init__(self, config_file=None, config_dict={}, initialize_site_data=None, log_level='info'):

        # start time for model run
        self.start_time = time.time()

        # initialize console handler for logger
        self.console_handler(log_level)

        logging.info("Starting CERF model")

        # inherit the configuration reader class attributes
        super(Model, self).__init__(config_file, config_dict)

        # siting data to use as the initial condition
        self.initialize_site_data = initialize_site_data

    def stage(self):
        """run model."""

        # prepare data for use in siting an expansion per state for a target year
        logging.info('Staging data...')

        # initial time for staging data
        staging_t0 = time.time()

        # prepare all data for state level run
        data = Stage(self.settings_dict,
                     self.lmp_zone_dict,
                     self.technology_dict,
                     self.technology_order,
                     self.infrastructure_dict,
                     self.initialize_site_data)

        logging.info(f'Staged data in {round((time.time() - staging_t0), 7)} seconds')

        return data

    def run_single_state(self, target_state_name, write_output=True):
        """run a single state."""

        # prepare all data for state level run
        data = self.stage()

        process = process_state(target_state_name=target_state_name,
                                settings_dict=self.settings_dict,
                                technology_dict=self.technology_dict,
                                technology_order=self.technology_order,
                                expansion_dict=self.expansion_dict,
                                states_dict=self.states_dict,
                                suitability_arr=data.suitability_arr,
                                lmp_arr=data.lmp_arr,
                                nov_arr=data.nov_arr,
                                ic_arr=data.ic_arr,
                                nlc_arr=data.nlc_arr,
                                zones_arr=data.zones_arr,
                                xcoords=data.xcoords,
                                ycoords=data.ycoords,
                                indices_2d=data.indices_2d,
                                randomize=self.settings_dict.get('randomize', True),
                                seed_value=self.settings_dict.get('seed_value', 0),
                                verbose=self.settings_dict.get('verbose', False),
                                write_output=write_output)

        logging.info(f"CERF model run completed in {round(time.time() - self.start_time, 7)} seconds")

        self.close_logger()

        return process



import calendar

import numpy as np


class NetOperationalValue:
    """Calculate Net Operational Value (NOV) in ($ / yr) per grid cell for all technologies.

    :param discount_rate:                   The time value of money in real terms.
                                            Units:  fraction
    :type discount_rate:                    float

    :param lifetime:                        Years of the expected technology plant lifetime.
                                            Units:  years
    :type lifetime:                         int

    :param unit_size:                       The size of the expected power plant.
                                            Units:  megawatt
    :type unit_size:                        int

    :param capacity_factor:                 Capacity factor defined as average annual power generated divided by the
                                            potential output if the plant operated at its rated capacity for a year.
                                            Units:  fraction
    :type capacity_factor:                  float

    :param variable_cost_esc_rate:          Escalation rate of variable cost.
                                            Units:  fraction
    :type variable_cost_esc_rate:           float

    :param fuel_esc_rate:                   Escalation rate of fuel.
                                            Units:  fraction
    :type fuel_esc_rate:                    float

    :param carbon_esc_rate:                 Escalation rate of carbon.
                                            Units:  fraction
    :type carbon_esc_rate:                  float

    :param variable_om:                     Variable operation and maintenance costs of yearly capacity use.
                                            Units:  $/MWh
    :type variable_om:                      float

    :param heat_rate:                       Amount of energy used by a power plant to generate one kilowatt-hour of
                                            electricity.
                                            Units:  Btu/kWh
    :type heat_rate:                        float

    :param fuel_price:                      Cost of fuel per unit.
                                            Units:  from GCAM ($/GJ) gets converted to ($/MBtu)
    :type fuel_price:                       float

    :param carbon_tax:                      The fee imposed on the burning of carbon-based fuels.
                                            Units:  $/ton
    :type carbon_tax:                       float

    :param carbon_capture_rate:             Rate of carbon capture.
                                            Units:  fraction
    :type carbon_capture_rate:              float

    :param fuel_co2_content:                CO2 content of the fuel and the heat rate of the technology.
                                            Units:  from GCAM (tons/MWh) gets converted to (tons/Btu)
    :type fuel_co2_content:                 float

    :param lmp_arr:                         Locational Marginal Price (LMP) per grid cell for each technology in a
                                            multi-dimensional array where the shape is [tech_id, xcoord, ycoord].
                                            Units:  $/MWh
    :type lmp_arr:                          ndarray

    :param target_year:                     Target year of the simulation as a four digit integer (e.g., 2010)
    :type target_year:                      int

    :param consider_leap_year:              Choose to account for leap year in the number of hours per year calculation
    :type consider_leap_year:               bool

    """

    # type hints
    discount_rate: float
    lifetime: int
    unit_size: int
    capacity_factor: float
    variable_cost_esc_rate: float
    fuel_esc_rate: float
    carbon_esc_rate: float
    variable_om: float
    heat_rate: float
    fuel_price: float
    carbon_tax: float
    carbon_capture_rate: float
    fuel_co2_content: float
    lmp_arr: np.ndarray
    target_year: int
    consider_leap_year: bool

    # constants for conversion
    FUEL_CO2_CONTENT_CONVERSION_FACTOR = 0.000000293071
    FUEL_PRICE_CONVERSION_FACTOR = 1.055056
    HOURS_PER_YEAR_NONLEAP = 8760
    HOURS_PER_YEAR_LEAP = 8784

    def __init__(self, discount_rate, lifetime, unit_size, capacity_factor, variable_cost_esc_rate,
                 fuel_esc_rate, carbon_esc_rate, variable_om, heat_rate, fuel_price, carbon_tax,
                 carbon_capture_rate, fuel_co2_content, lmp_arr, target_year, consider_leap_year=False):

        # assign class attributes
        self.discount_rate = discount_rate
        self.lifetime = lifetime
        self.unit_size = unit_size
        self.capacity_factor = capacity_factor
        self.variable_cost_esc = variable_cost_esc_rate
        self.fuel_esc = fuel_esc_rate
        self.carbon_esc = carbon_esc_rate
        self.variable_om = variable_om
        self.heat_rate = heat_rate
        self.carbon_tax = carbon_tax
        self.carbon_capture_rate = carbon_capture_rate
        self.lmp_arr = lmp_arr

        # create conversions
        self.fuel_price = self.convert_fuel_price(fuel_price)
        self.fuel_co2_content = self.convert_fuel_co2_content(fuel_co2_content)

        # adjust by leap year if so desired
        self.hours_per_year = self.assign_hours_per_year(target_year, consider_leap_year)

        # calculate annuity factor
        self.annuity_factor = self.calc_annuity_factor()

        # calculate levelizing factor for variable OM
        self.lf_vom = self.calc_levelization_factor_vom()

        # calculate levelizing factor for fuel
        self.lf_fuel = self.calc_levelization_factor_fuel()

        # calculate levelizing factor for carbon
        self.lf_carbon = self.calc_levelization_factor_carbon()

    @staticmethod
    def convert_fuel_price(fuel_price):
        """Convert fuel price from units $/GJ to $/MBtu"""

        return fuel_price * NetOperationalValue.FUEL_PRICE_CONVERSION_FACTOR

    @staticmethod
    def convert_fuel_co2_content(fuel_co2_content):
        """Convert fuel CO2 content from units tons/MWh to tons/Btu"""

        return fuel_co2_content * NetOperationalValue.FUEL_CO2_CONTENT_CONVERSION_FACTOR

    @classmethod
    def assign_hours_per_year(cls, target_year, consider_leap_year):
        """Assign the hours per year based on whether or not the target year is a leap year."""

        if calendar.isleap(target_year) and consider_leap_year:
            return cls.HOURS_PER_YEAR_LEAP
        else:
            return cls.HOURS_PER_YEAR_NONLEAP

    def calc_annuity_factor(self):
        """Calculate annuity factor."""

        fx = pow(1.0 + self.discount_rate, self.lifetime)
        return self.discount_rate * fx / (fx - 1.0)

    def calc_levelization_factor_vom(self):
        """Calculate the levelizing factor for variable OM."""

        k = (1.0 + self.variable_cost_esc) / (1.0 + self.discount_rate)
        return k * (1.0 - pow(k, self.lifetime)) * self.annuity_factor / (1.0 - k)

    def calc_levelization_factor_fuel(self):
        """Calculate the levelizing factor for fuel."""

        k = (1.0 + self.fuel_esc) / (1.0 + self.discount_rate)
        return k * (1.0 - pow(k, self.lifetime)) * self.annuity_factor / (1.0 - k)

    def calc_levelization_factor_carbon(self):
        """Calculate the levelizing factor for carbon."""

        k = (1.0 + self.carbon_esc) / (1.0 + self.discount_rate)
        return k * (1.0 - pow(k, self.lifetime)) * self.annuity_factor / (1.0 - k)

    def calc_generation(self):
        """Calculate electricity generation."""

        return self.unit_size * self.capacity_factor * self.hours_per_year

    def calc_nov(self):
        """Calculate NOV array for all technologies."""

        generation = self.calc_generation()
        term2 = self.lmp_arr * self.lf_fuel
        term3 = self.variable_om * self.lf_vom
        term4 = self.heat_rate * (self.fuel_price / 1000) * self.lf_fuel
        term5 = (self.carbon_tax * self.fuel_co2_content * self.heat_rate * self.lf_carbon / 1000000) * (1 - self.carbon_capture_rate)

        return generation * (term2 - (term3 + term4 + term5))
